"""This module contains classes and functions related to WhatsApp message templates."""

from __future__ import annotations

import dataclasses
import datetime
from typing import TYPE_CHECKING, cast

from pywa.listeners import TemplateUpdateListenerIdentifier
from pywa.types import CallbackData
from pywa.types.others import _ItemFactory
from pywa.types.templates import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.templates import (
    TemplateDetails as _TemplateDetails,
    CreatedTemplate as _CreatedTemplate,
    CreatedTemplates as _CreatedTemplates,
    UpdatedTemplate as _UpdatedTemplate,
    TemplateStatusUpdate as _TemplateStatusUpdate,
    TemplateQualityUpdate as _TemplateQualityUpdate,
    TemplateCategoryUpdate as _TemplateCategoryUpdate,
    TemplateComponentsUpdate as _TemplateComponentsUpdate,
)  # noqa MUST BE IMPORTED FIRST
from .others import Result, SuccessResult
from .. import filters as pywa_filters

if TYPE_CHECKING:
    from ..client import WhatsApp as WhatsAppAsync, WhatsApp
    from .sent_update import SentTemplate


@dataclasses.dataclass(frozen=True, kw_only=True)
class BaseTemplateUpdateAsync:
    """Base class for template updates."""

    _client: WhatsAppAsync = dataclasses.field(repr=False, hash=False, compare=False)
    template_id: str

    async def get_template(self) -> TemplateDetails:
        """
        Fetches the template details from WhatsApp.

        Returns:
            TemplateDetails: The details of the template.
        """
        return await self._client.get_template(
            template_id=self.template_id,
        )


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class TemplateStatusUpdate(BaseTemplateUpdateAsync, _TemplateStatusUpdate):
    """
    Represents status change of a template.

    Triggers::

        - A template is approved.
        - A template is rejected.
        - A template is disabled.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/reference/message_template_status_update>`_.

    Attributes:
        id: ID of Whatsapp Business Accounts this update belongs to.
        timestamp: Timestamp of the update (in UTC).
        new_status: The new status of the template.
        template_id: The ID of the template.
        template_name: The name of the template.
        template_language: The language of the template.
        reason: The reason the template was rejected (if status is ``REJECTED``).
        disable_date: The date the template was disabled (if status is ``DISABLED``).
        title: Title of template pause or unpause event.
        description: String describing why the template was locked or unlocked.
        shared_data: Shared data between handlers.
    """

    async def unpause(self) -> TemplateUnpauseResult:
        """
        Unpause the template that has been paused due to pacing.

        - You must wait 5 minutes after a template has been paused as a result of pacing before calling this method.
        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/message-templates/guidelines#unpausing>`_.

        Returns:
            A TemplateUnpauseResult object containing the result of the unpause operation.
        """
        return await self._client.unpause_template(template_id=self.template_id)


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class TemplateCategoryUpdate(BaseTemplateUpdateAsync, _TemplateCategoryUpdate):
    """
    Represents a template category update.

    Triggers::

        - The existing category of a WhatsApp template is going to be changed by an automated process.
        - The existing category of a WhatsApp template is changed manually or by an automated process.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/reference/template_category_update>`_.

    Attributes:
        id: ID of Whatsapp Business Accounts this update belongs to.
        timestamp: Timestamp of the update (in UTC).
        template_id: The ID of the template.
        template_name: The name of the template.
        template_language: The language of the template.
        correct_category: The category that the template will be `recategorized <https://developers.facebook.com/docs/whatsapp/updates-to-pricing/new-template-guidelines/#how-we-update-a-template-s-category-after-initial-approval>`_ as in 24 hours.
        previous_category: The template's previous category.
        new_category: If ``correct_category`` is set - the ``new_category`` is the current category of the template. If ``previous_category`` is set - the ``new_category`` is the category that the template was recategorized to.
    """


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class TemplateComponentsUpdate(BaseTemplateUpdateAsync, _TemplateComponentsUpdate):
    """
    Represents a template components update.

    Triggers::

        - A template is edited.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/reference/message_template_components_update>`_.

    Attributes:
        id: ID of Whatsapp Business Accounts this update belongs to.
        timestamp: Timestamp of the update (in UTC).
        template_id: The ID of the template.
        template_name: The name of the template.
        template_language: The language of the template.
        template_element: The body text of the template.
        template_title: The header text of the template.
        template_footer: The footer text of the template.
        template_buttons: A list of buttons in the template.
    """


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class TemplateQualityUpdate(BaseTemplateUpdateAsync, _TemplateQualityUpdate):
    """
    Represents a template quality update.

    Triggers::

        - A template's quality score changes.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/reference/message_template_quality_update>`_.

    Attributes:
        id: ID of Whatsapp Business Accounts this update belongs to.
        timestamp: Timestamp of the update (in UTC).
        template_id: The ID of the template.
        template_name: The name of the template.
        template_language: The language of the template.
        new_quality_score: The new quality score of the template.
        previous_quality_score: The previous quality score of the template.
    """


@dataclasses.dataclass(kw_only=True, slots=True)
class TemplateDetails(_TemplateDetails):
    """
    Represents the details of an existing WhatsApp Template.

    Attributes:
        id: The unique identifier of the template.
        name: The name of the template.
        status: The status of the template (See `Template Status <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#template-status>`_).
        category: The category of the template (See `Template Categorization <https://developers.facebook.com/docs/whatsapp/updates-to-pricing/new-template-guidelines#template-categorization>`_).
        language: The language of the template (See `Supported Languages <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/supported-languages>`_).
        components: Components that make up the template. Header, Body, Footer, Buttons, Cards, etc. (See `Template Components <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/components>`_).
        parameter_format: The type of parameter formatting the Header and BodyText components of the template will use.
        message_send_ttl_seconds: The time-to-live (TTL) for the template message in seconds (See `Time-to-live (TTL) <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#time-to-live--ttl---customization--defaults--min-max-values--and-compatibility>`_).
        correct_category: The correct category of the template, if applicable.
        previous_category: The previous category of the template, if applicable.
        rejected_reason: The reason the message template was rejected, if applicable (See `Template Rejected Status <https://developers.facebook.com/docs/whatsapp/updates-to-pricing/new-template-guidelines#rejected-status>`_).
        library_template_name: Template Library name that this template is cloned from, if applicable.
        quality_score: The quality score of the template, if applicable (See `Template Quality Score <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#template-quality-score>`_).
        cta_url_link_tracking_opted_out: Optional boolean field for opting out/in of link tracking at template level.
        sub_category: The sub-category of the template, if applicable.
    """

    _client: WhatsAppAsync = dataclasses.field(repr=False, hash=False, compare=False)

    async def delete(self) -> SuccessResult:
        """
        Delete this template

        - If you delete a template that has been sent in a template message but has yet to be delivered (e.g. because the customer's phone is turned off), the template's status will be set to ``PENDING_DELETION`` and we will attempt to deliver the message for 30 days. After this time you will receive a "Structure Unavailable" error and the customer will not receive the message.
        - Names of an approved template that has been deleted cannot be used again for 30 days.
        """
        return await self._client.delete_template(
            template_name=self.name, template_id=self.id
        )

    async def update(
        self,
        *,
        new_category: TemplateCategory | None = None,
        new_components: list[TemplateBaseComponent] | None = None,
        new_message_send_ttl_seconds: int | None = None,
        new_parameter_format: ParamFormat | None = None,
    ) -> UpdatedTemplate:
        """
        Update this template.

        - The template object will be updated in memory after a successful update.
        - Only templates with an ``APPROVED``, ``REJECTED``, or ``PAUSED`` status can be edited.
        - You cannot edit the category of an approved template.
        - Approved templates can be edited up to 10 times in a 30 day window, or 1 time in a 24 hour window. Rejected or paused templates can be edited an unlimited number of times.
        - After editing an approved or paused template, it will automatically be approved unless it fails template review.
        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#edit-a-message-template>`_.

        Args:
            new_category: The new category of the template (optional, cannot be changed for approved templates).
            new_components: The new components of the template (optional, if not provided, the existing components will be used).
            new_message_send_ttl_seconds: The new message send TTL in seconds (optional, if not provided, the existing TTL will be used).
            new_parameter_format: The new parameter format (optional, if not provided, the existing format will be used).

        Returns:
            Whether the template was updated successfully.
        """
        if res := await self._client.update_template(
            template_id=self.id,
            new_category=new_category,
            new_components=new_components,
            new_message_send_ttl_seconds=new_message_send_ttl_seconds,
            new_parameter_format=new_parameter_format,
        ):
            if new_category:
                self.category = new_category
            if new_components:
                self.components = new_components
            if new_message_send_ttl_seconds is not None:
                self.message_send_ttl_seconds = new_message_send_ttl_seconds
            if new_parameter_format is not None:
                self.parameter_format = new_parameter_format
        return res

    async def duplicate(self, **overrides) -> CreatedTemplate:
        """
        Duplicate this template.

        - WhatsApp Cloud API does not support duplicating templates, this method creates a new template with the same components and properties as this one. It is useful for creating variations of existing templates with minor changes.

        Example:
            >>> wa = WhatsApp(...)
            >>> template = wa.get_template("my_template_id")
            >>> new_template = template.duplicate(language=TemplateLanguage.ENGLISH)

        Args:
            overrides: Optional overrides for the template properties.
        """
        return await self._client.create_template(
            Template(
                name=overrides.get("name", self.name),
                language=overrides.get("language", self.language),
                category=overrides.get("category", self.category),
                components=overrides.get("components", self.components),
                parameter_format=overrides.get(
                    "parameter_format", self.parameter_format
                ),
                message_send_ttl_seconds=overrides.get(
                    "message_send_ttl_seconds", self.message_send_ttl_seconds
                ),
            )
        )

    async def compare(
        self,
        *to: int | str,
        start: datetime.datetime | int,
        end: datetime.datetime | int,
    ) -> TemplatesCompareResult:
        """
        You can compare two templates by examining how often each one is sent, which one has the lower ratio of blocks to sends, and each template's top reason for being blocked.

        - Only two templates can be compared at a time.
        - Both templates must be in the same WhatsApp Business Account.
        - Templates must have been sent at least 1,000 times in the queries specified timeframe.
        - Timeframes are limited to ``7``, ``30``, ``60`` and ``90`` day lookbacks from the time of the request.
        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/template-comparison>`_.

        Args:
            to: The IDs of the templates to compare with the given template.
            start: The start date of the comparison period.
            end: The end date of the comparison period.

        Returns:
            A TemplatesCompareResult object containing the comparison results.
        """
        return await self._client.compare_templates(
            template_id=self.id, *to, start=start, end=end
        )

    async def unpause(self) -> TemplateUnpauseResult:
        """
        Unpause a template that has been paused due to pacing.

        - This method can only be called if the template is currently paused due to pacing.
        - You must wait 5 minutes after a template has been paused as a result of pacing before calling this method.
        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/message-templates/guidelines#unpausing>`_.

        Returns:
            A TemplateUnpauseResult object containing the result of the unpause operation.
        """
        res = await self._client.unpause_template(template_id=self.id)
        if res:
            self.status = TemplateStatus.APPROVED
        return res

    async def send(
        self,
        to: str | int,
        params: list[TemplateBaseComponent.Params],
        *,
        reply_to_message_id: str | None = None,
        tracker: str | CallbackData | None = None,
        sender: str | int | None = None,
    ) -> SentTemplate:
        """
        A shortcut to send a template message to WhatsApp users.

        Args:
            to: The phone ID of the WhatsApp user.
            params: The parameters to fill in the template.
            reply_to_message_id: The ID of the message to reply to (optional).
            tracker: A callback data to track the message (optional, can be a string or a :class:`CallbackData` object).
            sender: The phone ID to send the template from (optional, if not provided, the client's phone ID will be used).
        """
        return await self._client.send_template(
            to=to,
            name=self.name,
            language=self.language,
            params=params,
            reply_to_message_id=reply_to_message_id,
            tracker=tracker,
            sender=sender,
        )


class TemplatesResult(Result[TemplateDetails]):
    """
    Represents the result of a templates query.

    Attributes:
        total_count: The total number of message templates that belong to a WhatsApp Business Account.
        message_template_count: The current number of message templates that belong to the WhatsApp Business Account.
        message_template_limit: The maximum number of message templates that can belong to a WhatsApp Business Account.
        are_translations_complete: The status for template translations.
    """

    total_count: int
    message_template_count: int
    message_template_limit: int
    are_translations_complete: bool

    def __init__(
        self,
        wa: WhatsApp,
        response: dict,
        item_factory: _ItemFactory,
    ):
        super().__init__(
            wa=wa,
            response=response,
            item_factory=item_factory,
        )
        self.total_count = response["summary"]["total_count"]
        self.message_template_count = response["summary"]["message_template_count"]
        self.message_template_limit = response["summary"]["message_template_limit"]
        self.are_translations_complete = response["summary"][
            "are_translations_complete"
        ]

    def __repr__(self) -> str:
        return (
            f"TemplatesResult({self._data!r}, has_next={self.has_next!r}, has_previous={self.has_previous!r}, "
            f"total_count={self.total_count!r}, "
            f"message_template_count={self.message_template_count!r}, "
            f"message_template_limit={self.message_template_limit!r}, "
            f"are_translations_complete={self.are_translations_complete!r})"
        )


class _CreatedAndUpdatedTemplateActionsAsync:
    async def get(self) -> TemplateDetails:
        """
        Retrieve the details of the created or updated template.

        Returns:
            TemplateDetails: The details of the created or updated template.
        """
        return await self._client.get_template(template_id=self.id)

    async def wait_until_approved(
        self,
        *,
        cancelers: pywa_filters.Filter | None = None,
        timeout: float | None = None,
    ) -> TemplateStatusUpdate:
        """
        Wait until the template is approved.

        Example usage:

            >>> from pywa import WhatsApp, filters
            >>> wa = WhatsApp(...)
            >>> created_template = wa.create_template(...)
            >>> status = created_template.wait_until_approved(cancelers=filters.template_status & filters.template_status_rejected)
            >>> print(f"Template {created_template.id} is approved with status: {status.new_status}")


        Args:
            cancelers: A filter to cancel the waiting process.
            timeout: The maximum time to wait for the template to be approved.

        Returns:
            TemplateStatusUpdate: An update containing the status of the template once it is approved.
        """
        return cast(
            TemplateStatusUpdate,
            await self._client.listen(
                to=TemplateUpdateListenerIdentifier(
                    waba_id=self._client.business_account_id, template_id=self.id
                ),
                filters=pywa_filters.template_status
                & pywa_filters.template_status_approved,
                cancelers=cancelers,
                timeout=timeout,
            ),
        )


@dataclasses.dataclass(frozen=True, slots=True)
class CreatedTemplate(_CreatedAndUpdatedTemplateActionsAsync, _CreatedTemplate):
    """
    Represents a created WhatsApp Template.

    Attributes:
        id: the template ID.
        status: the template status.
        category: the template category.
    """

    _client: WhatsAppAsync = dataclasses.field(repr=False, hash=False, compare=False)


@dataclasses.dataclass(frozen=True, slots=True)
class UpdatedTemplate(_CreatedAndUpdatedTemplateActionsAsync, _UpdatedTemplate):
    """
    Represents the result of a template update operation.

    Attributes:
        id: The unique identifier of the updated template.
        name: The name of the updated template.
        category: The category of the updated template.
        success: Indicates whether the template update was successful.
    """

    _client: WhatsAppAsync = dataclasses.field(repr=False, hash=False, compare=False)


@dataclasses.dataclass(frozen=True, slots=True)
class CreatedTemplates(_CreatedTemplates):
    """
    Represents a collection of created WhatsApp Templates.

    Attributes:
        templates: A list of CreatedTemplate instances.
    """

    templates: tuple[CreatedTemplate, ...]



from __future__ import annotations

from pywa.listeners import *  # noqa MUST BE IMPORTED FIRST
from pywa.listeners import (
    Listener as _Listener,
    BaseListenerIdentifier,
)  # noqa MUST BE IMPORTED FIRST

import asyncio
from typing import TYPE_CHECKING, TypeAlias

from pywa import utils, _helpers as helpers
from pywa.types.base_update import BaseUpdate
from .filters import Filter
from .types import (
    Message,
    CallbackButton,
    CallbackSelection,
    MessageStatus,
    ChatOpened,
    FlowCompletion,
)

if TYPE_CHECKING:
    from .client import WhatsApp


class Listener(_Listener):
    _listener_canceled = ListenerCanceled

    def __init__(
        self,
        wa: WhatsApp,
        identifier: BaseListenerIdentifier,
        filters: Filter,
        cancelers: Filter,
    ):
        self.filters = filters
        self.cancelers = cancelers
        self.future: asyncio.Future[BaseUpdate] = asyncio.Future()
        self.future.add_done_callback(
            lambda _: wa._remove_listener(identifier=identifier)
        )

    def set_result(self, result: BaseUpdate) -> None:
        self.future.set_result(result)

    def set_exception(self, exception: Exception) -> None:
        self.future.set_exception(exception)

    def is_set(self) -> bool:
        return self.future.done()

    async def apply_filters(self, wa: WhatsApp, update: BaseUpdate) -> bool:
        return not self.filters or await self.filters.check_async(wa, update)

    async def apply_cancelers(self, wa: WhatsApp, update: BaseUpdate) -> bool:
        return self.cancelers and await self.cancelers.check_async(wa, update)


class _AsyncListeners:
    async def listen(
        self: WhatsApp,
        to: BaseListenerIdentifier,
        *,
        filters: Filter = None,
        cancelers: Filter = None,
        timeout: float | None = None,
    ) -> BaseUpdate:
        """
        Listen to an update.

        - You can use one of the shortcuts to listen to a specific update type:
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_for_reply`
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_for_click`
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_for_selection`
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_for_completion`
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_until_read`
            - :meth:`~pywa_async.types.sent_message.SentMessage.wait_until_delivered`

        Example:

            .. code-block:: python

                    try:
                        await wa.send_message(
                            to="123456",
                            text="Send me a message",
                            buttons=[Button(title="Cancel", callback_data="cancel")]
                        )
                        update: Message = wa.listen(
                            to=UserUpdateListenerIdentifier(sender="123456", recipient="654321"),
                            filters=filters.message & filters.text,
                            cancelers=filters.callback_button & filters.matches("cancel"),
                            timeout=10
                        )
                        print(update)
                    except ListenerTimeout:
                        print("Listener timed out")
                    except ListenerCanceled:
                        print("Listener was canceled")
                    except ListenerStopped:
                        print("Listener was stopped")

        Args:
            to: The identifier of the update to listen to.
            filters: The filters to apply to the update, return the update if the filters pass.
            cancelers: The filters to cancel the listening, raise :class:`ListenerCanceled` if the update matches.
            timeout: The time to wait for the update, raise :class:`ListenerTimeout` if the time passes

        Returns:
            The update that passed the filters

        Raises:
            ListenerTimeout: If the listener timed out
            ListenerCanceled: If the listener was canceled by a filter
            ListenerStopped: If the listener was stopped manually
        """
        if self._server is utils.MISSING:
            raise ValueError(
                "You must initialize the WhatsApp client with an web app"
                " (Flask or FastAPI or custom server by setting `server` to None) in order to listen to incoming updates."
            )
        listener = Listener(
            wa=self,
            identifier=to,
            filters=filters,
            cancelers=cancelers,
        )
        self._listeners[to] = listener
        try:
            return await asyncio.wait_for(listener.future, timeout=timeout)
        except asyncio.TimeoutError:
            raise ListenerTimeout(timeout)



import dataclasses
import datetime
import importlib
import json
import pathlib
from typing import Callable

import pytest

from pywa import WhatsApp, handlers, utils, filters
from pywa.handlers import FlowCompletionHandler
from pywa.types.flows import (
    FlowJSON,
    Screen,
    Layout,
    Form,
    TextInput,
    InputType,
    FlowActionType,
    DataSource,
    ScreenDataRef,
    ComponentRef,
    ScreenData,
    FlowResponse,
    FlowRequest,
    FlowRequestActionType,
    _FlowJSONEncoder,
    Ref,
    NavigationItemEnd,
    NavigationItemMainContent,
    NavigationItem,
    NavigateAction,
    Next,
    FlowStr,
    FlowPreview,
)
from pywa.utils import Version


def test_flows_to_json():
    for version in pathlib.Path("tests/data/flows").iterdir():
        if version.is_dir():
            with open(
                pathlib.Path(version, "examples.json"), "r", encoding="utf-8"
            ) as f:
                json_examples = json.load(f)
                obj_examples = importlib.import_module(
                    f"tests.data.flows.{version.name}.examples"
                )
                for flow_name, flow_json in json_examples.items():
                    obj_dict = json.loads(getattr(obj_examples, flow_name).to_json())
                    example_dict = json_examples[flow_name]
                    assert obj_dict["version"] == version.name.replace("_", ".")
                    assert example_dict["version"] == version.name.replace("_", ".")
                    assert obj_dict == example_dict, (
                        f"Flow {version.name=} {flow_name=} does not match example"
                    )


def test_min_version():
    with pytest.raises(ValueError):
        FlowJSON(version="1.0", screens=[])


def test_empty_form():
    with pytest.raises(ValueError):
        Form(name="form", children=[])


def test_component_ref():
    assert ComponentRef("test").to_str() == "${form.test}"
    assert ComponentRef("test", screen="START").to_str() == "${screen.START.form.test}"
    assert TextInput(name="test", label="Test").ref.to_str() == "${form.test}"
    assert (
        TextInput(name="test", label="Test").ref_in(screen="START").to_str()
        == "${screen.START.form.test}"
    )
    in_ref = (
        Screen(id="START", layout=Layout(children=[]))
        / TextInput(name="test", label="Test").ref
    )
    assert in_ref.to_str() == "${screen.START.form.test}"
    assert isinstance(in_ref, ComponentRef)


def test_screen_data_ref():
    assert ScreenDataRef("test").to_str() == "${data.test}"
    assert ScreenDataRef("test", screen="START").to_str() == "${screen.START.data.test}"
    assert ScreenData(key="test", example="Example").ref.to_str() == "${data.test}"
    assert (
        ScreenData(key="test", example="Example").ref_in(screen="START").to_str()
        == "${screen.START.data.test}"
    )
    in_ref = (
        Screen(id="START", layout=Layout(children=[]))
        / ScreenData(key="test", example="Example").ref
    )
    assert in_ref.to_str() == "${screen.START.data.test}"
    assert isinstance(in_ref, ScreenDataRef)


def test_ref_to_str_without_screen():
    ref = Ref(prefix="data", field="age")
    assert ref.to_str() == "${data.age}"


def test_ref_to_str_with_screen_id():
    ref = Ref(prefix="data", field="age", screen="START")
    assert ref.to_str() == "${screen.START.data.age}"


def test_ref_to_str_with_screen():
    screen = Screen(id="START", layout=Layout(children=[]))
    ref = Ref(prefix="data", field="age", screen=screen)
    assert ref.to_str() == "${screen.START.data.age}"


def test_ref_equality():
    ref = Ref(prefix="data", field="age")
    condition = ref == 21
    assert condition.to_str() == "(${data.age} == 21)"


def test_ref_inequality():
    ref = Ref(prefix="data", field="age")
    condition = ref != 18
    assert condition.to_str() == "(${data.age} != 18)"


def test_ref_greater_than():
    ref = Ref(prefix="data", field="age")
    condition = ref > 21
    assert condition.to_str() == "(${data.age} > 21)"


def test_ref_greater_than_or_equal():
    ref = Ref(prefix="data", field="age")
    condition = ref >= 21
    assert condition.to_str() == "(${data.age} >= 21)"


def test_ref_less_than():
    ref = Ref(prefix="data", field="age")
    condition = ref < 21
    assert condition.to_str() == "(${data.age} < 21)"


def test_ref_less_than_or_equal():
    ref = Ref(prefix="data", field="age")
    condition = ref <= 21
    assert condition.to_str() == "(${data.age} <= 21)"


def test_ref_addition():
    ref = Ref(prefix="data", field="age")
    math_op = ref + 21
    assert math_op.to_str() == "(${data.age} + 21)"


def test_ref_right_addition():
    ref = Ref(prefix="data", field="age")
    math_op = 21 + ref
    assert math_op.to_str() == "(21 + ${data.age})"


def test_ref_subtraction():
    ref = Ref(prefix="data", field="age")
    math_op = ref - 21
    assert math_op.to_str() == "(${data.age} - 21)"


def test_ref_right_subtraction():
    ref = Ref(prefix="data", field="age")
    math_op = 21 - ref
    assert math_op.to_str() == "(21 - ${data.age})"


def test_ref_multiplication():
    ref = Ref(prefix="data", field="age")
    math_op = ref * 21
    assert math_op.to_str() == "(${data.age} * 21)"


def test_ref_right_multiplication():
    ref = Ref(prefix="data", field="age")
    math_op = 21 * ref
    assert math_op.to_str() == "(21 * ${data.age})"


def test_ref_division():
    ref = Ref(prefix="data", field="age")
    math_op = ref / 21
    assert math_op.to_str() == "(${data.age} / 21)"


def test_ref_right_division():
    ref = Ref(prefix="data", field="age")
    math_op = 21 / ref
    assert math_op.to_str() == "(21 / ${data.age})"


def test_ref_modulus():
    ref = Ref(prefix="data", field="age")
    math_op = ref % 21
    assert math_op.to_str() == "(${data.age} % 21)"


def test_ref_right_modulus():
    ref = Ref(prefix="data", field="age")
    math_op = 21 % ref
    assert math_op.to_str() == "(21 % ${data.age})"


def test_math_ref_with_ref():
    ref_one = Ref(prefix="data", field="age")
    ref_two = Ref(prefix="data", field="height")
    math_op = ref_one + ref_two
    assert math_op.to_str() == "(${data.age} + ${data.height})"


def test_ref_with_math_op():
    ref = Ref(prefix="data", field="age")
    math_op = 21 + ref
    ref_with_math_op = ref + math_op
    assert ref_with_math_op.to_str() == "(${data.age} + (21 + ${data.age}))"


def test_math_op_with_ref():
    ref = Ref(prefix="data", field="age")
    math_op = 21 + ref
    math_op_with_ref = math_op + ref
    assert math_op_with_ref.to_str() == "((21 + ${data.age}) + ${data.age})"


def test_math_op_with_math_op():
    ref = Ref(prefix="data", field="age")
    math_op_1 = 21 + ref
    math_op_2 = 21 + ref
    assert (
        math_op_1 + math_op_2
    ).to_str() == "((21 + ${data.age}) + (21 + ${data.age}))"


def test_math_op_with_number():
    ref = Ref(prefix="data", field="age")
    math_op = 21 + ref
    math_op_with_number = math_op + 21
    assert math_op_with_number.to_str() == "((21 + ${data.age}) + 21)"


def test_ref_with_math_op_and_ref():
    ref = Ref(prefix="data", field="age")
    math_op = 21 + ref
    math_op_with_ref = ref + math_op + ref
    assert (
        math_op_with_ref.to_str()
        == "((${data.age} + (21 + ${data.age})) + ${data.age})"
    )


def test_logical_and_with_ref():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition = ref1 & ref2
    assert condition.to_str() == "(${data.age} && ${form.is_verified})"


def test_logical_or_with_ref():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition = ref1 | ref2
    assert condition.to_str() == "(${data.age} || ${form.is_verified})"


def test_logical_and_with_condition():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition1 = ref1 > 21
    condition2 = ref2 == True  # noqa: E712
    combined_condition = condition1 & condition2
    assert (
        combined_condition.to_str()
        == "((${data.age} > 21) && (${form.is_verified} == true))"
    )


def test_logical_or_with_condition():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition1 = ref1 < 18
    condition2 = ref2 == False  # noqa:  E712
    combined_condition = condition1 | condition2
    assert (
        combined_condition.to_str()
        == "((${data.age} < 18) || (${form.is_verified} == false))"
    )


def test_invert_condition():
    ref = Ref(prefix="data", field="age")
    condition = ~ref
    assert condition.to_str() == "!${data.age}"


def test_combined_conditions_with_invert():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition = ~(ref1 > 18) & (ref2 == True)  # noqa: E712
    assert (
        condition.to_str() == "(!(${data.age} > 18) && (${form.is_verified} == true))"
    )


def test_combined_conditions_with_literal_before():
    ref1 = Ref(prefix="data", field="age")
    ref2 = Ref(prefix="form", field="is_verified")
    condition = ref2 & (ref1 > 18)
    assert condition.to_str() == "(${form.is_verified} && (${data.age} > 18))"


def test_combined_conditions_with_literal_after():
    ref1 = Ref(prefix="form", field="is_verified")
    ref2 = Ref(prefix="data", field="age")
    condition = (ref2 > 18) & ref1
    assert condition.to_str() == "((${data.age} > 18) && ${form.is_verified})"


def test_condition_with_backticks():
    ref = Ref(prefix="data", field="age")
    condition = ref > 18
    assert condition.to_str() == "(${data.age} > 18)"
    condition.wrap_with_backticks = True
    assert condition.to_str() == "`(${data.age} > 18)`"


def visible_conditions_wrapped_with_backticks():
    text = TextInput(name="test", label="Test", visible=Ref("data", "age") > 18)
    assert text.visible.to_str() == "`(${data.age} > 18)`"


def test_flow_str():
    assert (
        FlowStr(string="This is an example for Ana`s house.").to_str()
        == r"` 'This is an example for Ana\\`s house.' `"
    )

    assert (
        FlowStr(string="This is an example for Ana's house.").to_str()
        == r"` 'This is an example for Ana\\'s house.' `"
    )

    assert (
        FlowStr(
            string=r"This is an example with already escaped backticks \\`."
        ).to_str()
        == r"` 'This is an example with already escaped backticks \\`.' `"
    )

    assert (
        FlowStr(
            string="This is an example with vars {var1} and {var2}.",
            var1=ComponentRef("age"),
            var2=ComponentRef("height"),
        ).to_str()
        == r"` 'This is an example with vars ' ${form.age} ' and ' ${form.height} '.' `"
    )

    assert (
        FlowStr(
            string="This is an example with `{var1}` wrapped in backticks and '{var2}' with single quotes.",
            var1=ComponentRef("age"),
            var2=ComponentRef("height"),
        ).to_str()
        == r"` 'This is an example with \\`' ${form.age} '\\` wrapped in backticks and \\'' ${form.height} '\\' with single quotes.' `"
    )

    assert (
        FlowStr(
            string="This is an example with math expression {var1_plus_var2} and {var1_minus_var2}.",
            var1_plus_var2=ComponentRef("age") + ComponentRef("height"),
            var1_minus_var2=ComponentRef("age") - ComponentRef("height"),
        ).to_str()
        == r"` 'This is an example with math expression ' (${form.age} + ${form.height}) ' and ' (${form.age} - ${form.height}) '.' `"
    )


def test_init_values():
    text_entry = TextInput(name="test", label="Test", init_value="Example")
    form = Form(name="form", children=[text_entry])
    assert form.init_values == {"test": "Example"}

    # check for duplicate init_values (in the form level and in the children level)
    # with pytest.raises(ValueError):
    #     TextInput(
    #         name="test", label="Test", init_value="Example", input_type=InputType.NUMBER
    #     )
    #     Form(name="form", init_values={"test": "Example"}, children=[text_entry])

    # test that if form has init_values referred to a ref,
    # the init_values does not fill up from the .children init_value's
    form_with_init_values_as_data_key = Screen(
        id="test",
        title="Test",
        data=[
            init_vals := ScreenData(key="init_vals", example={"test": "Example"}),
        ],
        layout=Layout(
            children=[
                Form(name="form", init_values=init_vals.ref, children=[text_entry])
            ]
        ),
    )
    assert isinstance(
        form_with_init_values_as_data_key.layout.children[0].init_values, Ref
    )


#
#
def test_error_messages():
    text_entry = TextInput(name="test", label="Test", error_message="Example")
    form = Form(name="form", children=[text_entry])
    assert form.error_messages == {"test": "Example"}

    # check for duplicate error_messages (in the form level and in the children level)
    # with pytest.raises(ValueError):
    #     TextInput(name="test", label="Test", error_message="Example")
    #     Form(name="form", error_messages={"test": "Example"}, children=[text_entry])

    # test that if form has error_messages referred to a ref,
    # the error_messages does not fill up from the .children error_message's
    form_with_error_messages_as_data_key = Screen(
        id="test",
        title="Test",
        data=[
            error_msgs := ScreenData(key="error_msgs", example={"test": "Example"}),
        ],
        layout=Layout(
            children=[
                Form(
                    name="form",
                    error_messages=error_msgs.ref,
                    children=[text_entry],
                )
            ]
        ),
    )

    assert isinstance(
        form_with_error_messages_as_data_key.layout.children[0].error_messages, Ref
    )


def test_encoder_py_to_json_types():
    encoder = _FlowJSONEncoder()
    assert encoder._get_json_type("Example") == {"type": "string"}
    assert encoder._get_json_type(example=1) == {"type": "number"}
    assert encoder._get_json_type(1.0) == {"type": "number"}
    assert encoder._get_json_type(True) == {"type": "boolean"}


def test_encoder_single_data_source():
    encoder = _FlowJSONEncoder()
    assert encoder._get_json_type(DataSource(id="1", title="Example")) == {
        "type": "object",
        "properties": {"id": {"type": "string"}, "title": {"type": "string"}},
    }


def test_encoder_navigation_item():
    encoder = _FlowJSONEncoder()
    assert encoder._get_json_type(
        [
            NavigationItem(
                id="1",
                main_content=NavigationItemMainContent(
                    title="First item",
                    description="first item",
                    metadata="metadata",
                ),
                end=NavigationItemEnd(
                    title="End",
                    description="the end",
                ),
                on_click_action=NavigateAction(next=Next(name="SECOND_SCREEN")),
            )
        ]
    ) == {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "main-content": {
                    "type": "object",
                    "properties": {
                        "title": {"type": "string"},
                        "description": {"type": "string"},
                        "metadata": {"type": "string"},
                    },
                },
            },
        },
    }


def test_encoder_multiple_data_sources():
    encoder = _FlowJSONEncoder()
    assert encoder._get_json_type(
        [
            DataSource(id="1", title="Example"),
            DataSource(id="2", title="Example2"),
        ]
    ) == {
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "title": {"type": "string"},
            },
        },
    }


def test_screen_data_empty_example():
    with pytest.raises(ValueError):
        FlowJSON(
            screens=[
                Screen(
                    id="test",
                    title="Test",
                    data=[ScreenData(key="test", example=[])],
                    layout=Layout(children=[]),
                )
            ],
            version="2.1",
        ).to_json()


def test_screen_data_invalid_type():
    with pytest.raises(ValueError):
        FlowJSON(
            screens=[
                Screen(
                    id="test",
                    title="Test",
                    data=[ScreenData(key="test", example=ValueError)],
                    layout=Layout(children=[]),
                )
            ],
            version="2.1",
        ).to_json()


def test_flow_response_with_error_msg():
    assert (
        "error_message"
        in FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            screen="TEST",
            error_message="Example",
        ).to_dict()["data"]
    )


def test_flow_response_with_close_flow():
    assert (
        FlowResponse(  # closing flow make screen `SUCCESS`
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
            flow_token="test",
        ).to_dict()["screen"]
        == "SUCCESS"
    )

    with pytest.raises(ValueError):  # not closing flow without screen
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            flow_token="test",
            close_flow=False,
        )

    with pytest.raises(ValueError):  # closing flow without flow_token
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
        )

    with pytest.raises(ValueError):  # closing flow with error_message
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
            flow_token="fdf",
            error_message="Example",
        )


def test_flow_response_with_data_sources():
    assert FlowResponse(
        version=Version.FLOW_MSG.value,
        data={"data_source": DataSource(id="1", title="Example")},
        screen="TEST",
    ).to_dict()["data"]["data_source"] == {"id": "1", "title": "Example"}

    assert FlowResponse(
        version=Version.FLOW_MSG.value,
        data={"data_source": [DataSource(id="1", title="Example")]},
        screen="TEST",
    ).to_dict()["data"]["data_source"] == [{"id": "1", "title": "Example"}]


@pytest.fixture
def flow_request():
    return FlowRequest(
        version=...,
        action=FlowRequestActionType.DATA_EXCHANGE,
        flow_token="xyz",
        screen="START",
        data={},
        raw=...,
        raw_encrypted=...,
    )


def get_flow_callback_wrapper(callback: Callable):
    wa = WhatsApp(
        token="xxx", server=None, business_private_key="xxx", verify_token="fdfd"
    )
    wrapper = wa.get_flow_request_handler(
        endpoint="/flow",
        callback=callback,
        request_decryptor=...,
        response_encryptor=...,
    )
    return wrapper


def test_flow_callback_wrapper_main_handler(flow_request):
    def main_handler(_, __): ...

    wrapper = get_flow_callback_wrapper(main_handler)
    assert wrapper._get_callback(flow_request) is main_handler


def test_flow_callback_wrapper_screen(flow_request):
    def data_exchange_start_screen_callback(_, __): ...

    wrapper = get_flow_callback_wrapper(lambda _, __: ...)
    wrapper.add_handler(
        callback=data_exchange_start_screen_callback,
        action=FlowRequestActionType.DATA_EXCHANGE,
        screen="START",
    )
    req = dataclasses.replace(flow_request, screen="START")
    assert wrapper._get_callback(req) is data_exchange_start_screen_callback

    def data_exchange_callback_without_screen(_, __): ...

    wrapper.add_handler(
        callback=data_exchange_callback_without_screen,
        action=FlowRequestActionType.DATA_EXCHANGE,
        screen=None,
    )
    assert wrapper._get_callback(req) is data_exchange_callback_without_screen


def test_flow_callback_wrapper_filters(flow_request):
    def init_with_data_filter(_, __): ...

    wrapper = get_flow_callback_wrapper(lambda _, __: ...)
    wrapper.add_handler(
        callback=init_with_data_filter,
        action=FlowRequestActionType.INIT,
        screen=None,
        filters=filters.new(lambda _, r: r.data.get("age") >= 20),
    )
    req = dataclasses.replace(
        flow_request, action=FlowRequestActionType.INIT, data={"age": 20}
    )
    assert wrapper._get_callback(req) is init_with_data_filter


def test_flow_callback_wrapper_on_error(flow_request):
    wrapper = get_flow_callback_wrapper(lambda _, __: ...)

    @wrapper.on_data_exchange(call_on_error=False)
    def not_on_error(_, __): ...

    @wrapper.on_data_exchange(call_on_error=True)
    def on_error(_, __): ...

    req = dataclasses.replace(
        flow_request,
        action=FlowRequestActionType.DATA_EXCHANGE,
        data={"error_message": "Example"},
    )
    assert wrapper._get_callback(req) is on_error


def test_flow_callback_wrapper_on_completion():
    wrapper = get_flow_callback_wrapper(lambda _, __: ...)

    @wrapper.on_completion
    def on_completion(_, __): ...

    assert wrapper._wa._handlers[FlowCompletionHandler][0]._callback is on_completion


def test_flows_server():
    with pytest.raises(ValueError, match="^When using a custom server.*"):
        wa = WhatsApp(token=..., server=None, verify_token=...)
        wa.add_flow_request_handler(
            handlers.FlowRequestHandler(
                callback=...,
                endpoint=...,
            )
        )

    with pytest.raises(ValueError, match="^You must initialize the WhatsApp client.*"):
        wa = WhatsApp(token=..., server=utils.MISSING)
        wa.add_flow_request_handler(
            handlers.FlowRequestHandler(
                callback=...,
                endpoint=...,
            )
        )


def test_flow_preview_with_params():
    preview = FlowPreview(
        url="https://business.facebook.com/wa/manage/flows/1460367762010364/preview/?token=fihsufcisd-09ad-4b88-b6aa-hdiewfcw",
        expires_at=datetime.datetime.now(),
    )
    assert (
        preview.with_params(
            interactive=True,
            flow_token="flow_token_example",
            flow_action=FlowRequestActionType.NAVIGATE,
            flow_action_payload={
                "screen": "START",
                "is_new_user": True,
                "welcome_msg": "Welcome to our service!",
            },
            phone_number="1234567890",
            debug=True,
        )
        == "https://business.facebook.com/wa/manage/flows/1460367762010364/preview/?token=fihsufcisd-09ad-4b88-b6aa-hdiewfcw&flow_token=flow_token_example&interactive=true&flow_action=navigate&flow_action_payload=%7B%22screen%22%3A%22START%22%2C%22is_new_user%22%3Atrue%2C%22welcome_msg%22%3A%22Welcome+to+our+service%21%22%7D&phone_number=1234567890&debug=true"
    )



from __future__ import annotations

"""This module contains the types related to system messages."""

__all__ = ["PhoneNumberChange", "IdentityChange", "SystemType", "Identity"]


from .. import utils

import dataclasses
import datetime
from typing import TYPE_CHECKING

from .base_update import BaseUserUpdate  # noqa
from .others import (
    MessageType,
    Metadata,
)

if TYPE_CHECKING:
    from ..client import WhatsApp


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class PhoneNumberChange(BaseUserUpdate):
    """
    A update received when a user changes their phone number on WhatsApp.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The type of the message (always ``MessageType.SYSTEM``).
        sys_type: The type of the system message (always ``SystemType.USER_CHANGED_NUMBER``).
        from_user: The user who changed their phone number. THIS IS THE OLD WA ID!
        timestamp: The timestamp when the message was arrived to WhatsApp servers (in UTC).
        old_wa_id: The old WhatsApp ID of the user.
        new_wa_id: The new WhatsApp ID of the user.
        body: The body of the system message (e.g., `John changed their phone number`).
    """

    type: MessageType
    sys_type: SystemType
    old_wa_id: str
    new_wa_id: str
    body: str

    _webhook_field = "messages"

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> PhoneNumberChange:
        sys = (
            msg := (value := (entry := update["entry"][0])["changes"][0]["value"])[
                "messages"
            ][0]
        )["system"]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=msg["id"],
            timestamp=datetime.datetime.fromtimestamp(
                int(msg["timestamp"]),
                datetime.timezone.utc,
            ),
            metadata=Metadata.from_dict(value["metadata"]),
            type=MessageType(msg["type"]),
            sys_type=SystemType(sys["type"]),
            from_user=client._usr_cls(_client=client, wa_id=msg["from"], name=None),
            old_wa_id=sys.get("customer", msg["from"]),  # v12^ from
            new_wa_id=sys.get("new_wa_id", sys["wa_id"]),  # v12^ wa_id
            body=sys["body"],
        )


class SystemType(utils.StrEnum):
    """
    The type of the system message.

    Attributes:
        USER_CHANGED_NUMBER: A user changed their phone number.
        CUSTOMER_IDENTITY_CHANGED: A user changed their profile information.
    """

    USER_CHANGED_NUMBER = "user_changed_number"
    CUSTOMER_IDENTITY_CHANGED = "customer_identity_changed"

    UNKNOWN = "UNKNOWN"

    @classmethod
    def _missing_(cls, value: str):
        # the documentation very confusing about the values of system types
        if (
            value == "customer_changed_number"
        ):  # https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/components#messages-object
            return cls.USER_CHANGED_NUMBER  # https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/payload-examples#user-changed-number-notification
        return super()._missing_(value)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class IdentityChange(BaseUserUpdate):
    """
    A message received when a user changes their profile information on WhatsApp.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The type of the message (always ``MessageType.SYSTEM``).
        from_user: The user who changed their profile information.
        timestamp: The timestamp when the message was arrived to WhatsApp servers (in UTC).
        body: The body of the system message (e.g., `John changed their profile information`).
        identity: The new identity of the user (see :class:`Identity`).
    """

    type: MessageType
    sys_type: SystemType
    body: str
    identity: Identity

    _webhook_field = "messages"

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> IdentityChange:
        msg = (value := (entry := update["entry"][0])["changes"][0]["value"])[
            "messages"
        ][0]
        sys, identity = msg["system"], msg["identity"]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=msg["id"],
            type=MessageType(msg["type"]),
            sys_type=SystemType(sys["type"]),
            timestamp=datetime.datetime.fromtimestamp(
                int(msg["timestamp"]),
                datetime.timezone.utc,
            ),
            metadata=Metadata.from_dict(value["metadata"]),
            from_user=client._usr_cls.from_dict(value["contacts"][0], client=client),
            body=sys["body"],
            identity=Identity.from_dict(identity),
        )


@dataclasses.dataclass(frozen=True, slots=True)
class Identity:
    """
    Represents a user's identity on WhatsApp.

    Attributes:
        acknowledged: State of acknowledgment.
        created_timestamp: The time when the WhatsApp Business Management API detected the customer may have changed their profile information.
        hash: The hash of the identity.
    """

    acknowledged: bool
    created_timestamp: datetime.datetime
    hash: str

    @classmethod
    def from_dict(cls, data: dict) -> Identity:
        return cls(
            acknowledged=data["acknowledged"],
            created_timestamp=datetime.datetime.fromtimestamp(
                int(data["created_timestamp"]), datetime.timezone.utc
            ),
            hash=data["hash"],
        )



import pytest

from pywa.types.templates import *  # noqa: F403
import importlib
import json
import pathlib


def test_templates_to_json(caplog):
    for templates_dir in pathlib.Path("tests/data/templates").iterdir():
        with open(
            pathlib.Path(templates_dir, "examples.json"), "r", encoding="utf-8"
        ) as f:
            json_examples = json.load(f)
            obj_examples = importlib.import_module(
                f"tests.data.templates.{templates_dir.name}.examples"
            )
            for template_name, template_json in json_examples.items():
                example_obj: Template = getattr(obj_examples, template_name)
                example_dict: dict = json_examples[template_name]

                assert (
                    json.loads(
                        Template.from_dict(json.loads(example_obj.to_json())).to_json()
                    )
                    == example_dict
                ), (
                    f"Template {templates_dir.name=} {template_name=} does not match example JSON."
                )
                assert not caplog.records, (
                    f"Template {templates_dir.name=} {template_name=} has warnings: {' '.join(record.message for record in caplog.records)}"
                )


def test_template_text_examples_params():
    with pytest.raises(ValueError):
        HeaderText.Params("pos", named="named")


def test_template_text_examples_positionals():
    h = HeaderText(
        "Hi {{1}}, this is a text template number {{2}}",
        "David",
        1,
    )
    assert h.to_dict() == {
        "type": "HEADER",
        "format": "TEXT",
        "text": "Hi {{1}}, this is a text template number {{2}}",
        "example": {
            "header_text": ["David", "1"],
        },
    }
    assert h.text == "Hi {{1}}, this is a text template number {{2}}"
    assert h.example == ("David", 1)
    assert h.preview() == "Hi David, this is a text template number 1"
    assert h.params("John", 100).to_dict() == {
        "type": "HEADER",
        "parameters": [
            {"type": "text", "text": "John"},
            {"type": "text", "text": "100"},
        ],
    }
    assert h.param_format == ParamFormat.POSITIONAL
    with pytest.raises(
        ValueError,
        match="HeaderText does not support named parameters when text is positional.",
    ):
        h.params(named="John")
    with pytest.raises(
        ValueError,
        match="HeaderText requires 2 positional parameters, got 3.",
    ):
        h.params("John", 100, "unexpected")

    with pytest.raises(
        ValueError,
        match="HeaderText does not support parameters, as it has no example.",
    ):
        HeaderText("Hi").params("John")


def test_template_text_examples_named():
    b = BodyText(
        "Hi {{name}}, this is a text template number {{number}}",
        name="David",
        number=1,
    )
    assert b.to_dict() == {
        "type": "BODY",
        "text": "Hi {{name}}, this is a text template number {{number}}",
        "example": {
            "body_text_named_params": [
                {"param_name": "name", "example": "David"},
                {"param_name": "number", "example": "1"},
            ]
        },
    }
    assert b.text == "Hi {{name}}, this is a text template number {{number}}"
    assert b.example == {"name": "David", "number": 1}
    assert b.preview() == "Hi David, this is a text template number 1"
    assert b.params(name="John", number=100).to_dict() == {
        "type": "BODY",
        "parameters": [
            {"type": "text", "text": "John", "parameter_name": "name"},
            {"type": "text", "text": "100", "parameter_name": "number"},
        ],
    }
    assert b.param_format == ParamFormat.NAMED
    with pytest.raises(
        ValueError,
        match="BodyText does not support positional parameters when text is named.",
    ):
        b.params("John")
    with pytest.raises(
        ValueError,
        match="BodyText received unexpected parameters: unexpected",
    ):
        b.params(name="John", number=100, unexpected="unexpected")

    with pytest.raises(ValueError, match="BodyText is missing parameters: number"):
        b.params(name="John")

    with pytest.raises(
        ValueError,
        match="BodyText does not support parameters, as it has no example.",
    ):
        BodyText("Hi").params(name="John")


def test_components_that_does_not_get_params():
    with pytest.raises(ValueError):
        FooterText(
            text="This is a footer text template",
        ).params("John")


def test_media_header_handle_reset():
    media_header = HeaderImage(example="https://example.com/image.jpg")
    media_header._handle = "1:handle"
    media_header.example = "https://example2.com/image.jpg"
    assert media_header._handle is None, (
        "Handle should be reset to None after example change"
    )



import datetime

from pywa import types, WhatsApp
from pywa.types import Result
from pywa.types.flows import FlowDetails
from pywa.types.sent_update import (
    SentMessage,
    SentTemplate,
    SentTemplateStatus,
    InitiatedCall,
)
import pytest
from unittest.mock import MagicMock, call

wa = WhatsApp(phone_id="123456789", token="token")


def test_flow_details():
    assert FlowDetails.from_dict(
        client=wa,
        data={
            "id": "12345",
            "name": "my_flow",
            "status": "DRAFT",
            "updated_at": "2024-10-29T21:09:19+0000",
            "categories": ["OTHER"],
            "validation_errors": [],
            "json_version": "5.1",
            "data_api_version": "3.0",
            "endpoint_uri": "https://example.com/flow",
            "preview": {
                "preview_url": "https://business.facebook.com/wa/manage/flows/12345/preview/?token=xyz",
                "expires_at": "2024-12-31T18:45:52+0000",
            },
            "whatsapp_business_account": {
                "id": "124578",
                "name": "Test WhatsApp Business Account",
                "timezone_id": "1",
                "message_template_namespace": "1acaxyzfec",
            },
            "application": {
                "link": "https://www.facebook.com/games/?app_id=12345",
                "name": "My App",
                "id": "12345",
            },
        },
    ) == FlowDetails(
        _client=wa,
        id="12345",
        name="my_flow",
        status=types.flows.FlowStatus.DRAFT,
        updated_at=datetime.datetime(
            2024, 10, 29, 21, 9, 19, tzinfo=datetime.timezone.utc
        ),
        categories=(types.flows.FlowCategory.OTHER,),
        validation_errors=None,
        json_version="5.1",
        data_api_version="3.0",
        endpoint_uri="https://example.com/flow",
        preview=types.flows.FlowPreview(
            url="https://business.facebook.com/wa/manage/flows/12345/preview/?token=xyz",
            expires_at=datetime.datetime(
                2024, 12, 31, 18, 45, 52, tzinfo=datetime.timezone.utc
            ),
        ),
        whatsapp_business_account=types.others.WhatsAppBusinessAccount(
            id="124578",
            name="Test WhatsApp Business Account",
            timezone_id="1",
            message_template_namespace="1acaxyzfec",
            status=None,
            business_verification_status=None,
            is_enabled_for_insights=None,
            marketing_messages_onboarding_status=None,
            marketing_messages_lite_api_status=None,
            on_behalf_of_business_info=None,
            ownership_type=None,
            health_status=None,
            country=None,
            currency=None,
            subscribed_apps=None,
        ),
        application=types.flows.FacebookApplication(
            link="https://www.facebook.com/games/?app_id=12345",
            name="My App",
            id="12345",
        ),
    )


def test_business_profile():
    assert types.BusinessProfile.from_dict(
        data={
            "about": "Hey there! I am using PyWa.",
            "email": "user@server.com",
            "websites": ["https://pywa.readthedocs.io/"],
            "vertical": "PROF_SERVICES",
            "messaging_product": "whatsapp",
            "address": "123 Main St, San Francisco, CA 94107",
            "description": "Business Description",
        }
    ) == types.BusinessProfile(
        about="Hey there! I am using PyWa.",
        email="user@server.com",
        websites=("https://pywa.readthedocs.io/",),
        industry=types.others.Industry.PROF_SERVICES,
        profile_picture_url=None,
        description="Business Description",
        address="123 Main St, San Francisco, CA 94107",
    )


def test_business_phone_number():
    assert types.BusinessPhoneNumber.from_dict(
        data={
            "id": "123456789",
            "verified_name": "Test Number",
            "display_phone_number": "+1 123-456-7890",
            "conversational_automation": {
                "prompts": ["Hi man"],
                "commands": [
                    {"command_name": "start", "command_description": "Start"},
                    {"command_name": "help", "command_description": "Help"},
                ],
                "enable_welcome_message": False,
                "id": "47328947638",
            },
            "status": "CONNECTED",
            "quality_rating": "GREEN",
            "quality_score": {"score": "GREEN"},
            "webhook_configuration": {"application": "https://example.com/wa_webhook"},
            "name_status": "APPROVED",
            "new_name_status": "NONE",
            "code_verification_status": "NOT_VERIFIED",
            "account_mode": "LIVE",
            "is_on_biz_app": True,
            "is_official_business_account": False,
            "is_pin_enabled": True,
            "is_preverified_number": False,
            "messaging_limit_tier": "TIER_1K",
            "search_visibility": "NON_VISIBLE",
            "platform_type": "CLOUD_API",
            "throughput": {"level": "STANDARD"},
            "eligibility_for_api_business_global_search": "NON_ELIGIBLE_UNDEFINED_COUNTRY",
            "health_status": {
                "can_send_message": "BLOCKED",
                "entities": [
                    {
                        "entity_type": "PHONE_NUMBER",
                        "id": "123456789",
                        "can_send_message": "AVAILABLE",
                    },
                    {
                        "entity_type": "WABA",
                        "id": "8976r843r3r3",
                        "can_send_message": "BLOCKED",
                        "errors": [
                            {
                                "error_code": 141006,
                                "error_description": "There is an error with the payment method. This will block business initiated conversations.",
                                "possible_solution": "There was an error with your payment method. Please add a new payment method to the account.",
                            }
                        ],
                    },
                    {
                        "entity_type": "BUSINESS",
                        "id": "473958473543",
                        "can_send_message": "AVAILABLE",
                    },
                    {
                        "entity_type": "APP",
                        "id": "4759843753",
                        "can_send_message": "AVAILABLE",
                    },
                ],
            },
        }
    ) == types.BusinessPhoneNumber(
        id="123456789",
        verified_name="Test Number",
        display_phone_number="+1 123-456-7890",
        conversational_automation=types.others.ConversationalAutomation(
            id="47328947638",
            chat_opened_enabled=False,
            ice_breakers=("Hi man",),
            commands=(
                types.others.Command(name="start", description="Start"),
                types.others.Command(name="help", description="Help"),
            ),
        ),
        status="CONNECTED",
        quality_rating="GREEN",
        quality_score={"score": "GREEN"},
        webhook_configuration={"application": "https://example.com/wa_webhook"},
        name_status="APPROVED",
        new_name_status="NONE",
        code_verification_status="NOT_VERIFIED",
        account_mode="LIVE",
        is_on_biz_app=True,
        is_official_business_account=False,
        is_pin_enabled=True,
        is_preverified_number=False,
        messaging_limit_tier="TIER_1K",
        search_visibility="NON_VISIBLE",
        platform_type="CLOUD_API",
        throughput={"level": "STANDARD"},
        eligibility_for_api_business_global_search="NON_ELIGIBLE_UNDEFINED_COUNTRY",
        health_status={
            "can_send_message": "BLOCKED",
            "entities": [
                {
                    "entity_type": "PHONE_NUMBER",
                    "id": "123456789",
                    "can_send_message": "AVAILABLE",
                },
                {
                    "entity_type": "WABA",
                    "id": "8976r843r3r3",
                    "can_send_message": "BLOCKED",
                    "errors": [
                        {
                            "error_code": 141006,
                            "error_description": "There is an error with the payment method. This will block business initiated conversations.",
                            "possible_solution": "There was an error with your payment method. Please add a new payment method to the account.",
                        }
                    ],
                },
                {
                    "entity_type": "BUSINESS",
                    "id": "473958473543",
                    "can_send_message": "AVAILABLE",
                },
                {
                    "entity_type": "APP",
                    "id": "4759843753",
                    "can_send_message": "AVAILABLE",
                },
            ],
        },
        certificate=None,
        new_certificate=None,
        last_onboarded_time=None,
    )


def test_sent_message():
    sm = SentMessage.from_sent_update(
        client=wa,
        update={
            "messaging_product": "whatsapp",
            "contacts": [{"input": "16505555555", "wa_id": "16505555555"}],
            "messages": [
                {"id": "wamid.HBgLMTY1MDUwNzY1MjAVAgARGBI5QTNDQTVCM0Q0Q0Q2RTY3RTcA"}
            ],
        },
        from_phone_id=wa.phone_id,
    )
    assert sm == SentMessage(
        _client=wa,
        _callback_options=None,
        _flow_token=None,
        id="wamid.HBgLMTY1MDUwNzY1MjAVAgARGBI5QTNDQTVCM0Q0Q0Q2RTY3RTcA",
        to_user=types.User(wa_id="16505555555", name=None, _client=wa),
        from_phone_id=wa.phone_id,
    )
    assert sm.sender == wa.phone_id
    assert sm.recipient == sm.to_user.wa_id

    assert SentTemplate.from_sent_update(
        client=wa,
        update={
            "messaging_product": "whatsapp",
            "contacts": [{"input": "16505555555", "wa_id": "16505555555"}],
            "messages": [
                {
                    "id": "wamid.HBgLMTY1MDUwNzY1MjAVAgARGBI5QTNDQTVCM0Q0Q0Q2RTY3RTcA",
                    "message_status": "accepted",
                }
            ],
        },
        from_phone_id=wa.phone_id,
    ) == SentTemplate(
        _client=wa,
        _callback_options=None,
        _flow_token=None,
        id="wamid.HBgLMTY1MDUwNzY1MjAVAgARGBI5QTNDQTVCM0Q0Q0Q2RTY3RTcA",
        to_user=types.User(wa_id="16505555555", name=None, _client=wa),
        from_phone_id=wa.phone_id,
        status=SentTemplateStatus.ACCEPTED,
    )


def test_initiated_call():
    c = InitiatedCall.from_sent_update(
        client=wa,
        update={
            "messaging_product": "whatsapp",
            "calls": [{"id": "wacid.fiurefh8e="}],
            "success": True,
        },
        from_phone_id=wa.phone_id,
        to_wa_id="16506666666",
    )
    assert c.caller == wa.phone_id
    assert c.callee == "16506666666"
    assert c == InitiatedCall(
        _client=wa,
        id="wacid.fiurefh8e=",
        to_user=types.User(wa_id="16506666666", name=None, _client=wa),
        from_phone_id=wa.phone_id,
        success=True,
    )


@pytest.fixture
def fake_item_factory():
    return lambda data: {"name": data["name"]}


@pytest.fixture
def wa_result():
    _wa = MagicMock()
    _wa.api._make_request = MagicMock()
    return _wa


@pytest.fixture
def response_data():
    return {
        "data": [{"name": "Alice"}, {"name": "Bob"}],
        "paging": {
            "next": "/next-url",
            "previous": "/prev-url",
            "cursors": {"before": "abc", "after": "xyz"},
        },
    }


@pytest.fixture
def pages():
    return (
        {
            "data": [{"name": "Zed"}],
            "paging": {"next": "/page2", "previous": None, "cursors": {}},
        },
        {
            "data": [{"name": "Yara"}],
            "paging": {"next": "/page3", "previous": "/page1", "cursors": {}},
        },
        {
            "data": [{"name": "Xander"}],
            "paging": {"next": None, "previous": "/page2", "cursors": {}},
        },
    )


def test_result_behavior(wa_result, fake_item_factory, response_data):
    result = Result(wa_result, response_data, fake_item_factory)
    assert len(result) == 2
    assert list(result) == [{"name": "Alice"}, {"name": "Bob"}]
    assert result[0]["name"] == "Alice"
    assert bool(result)


def test_result_properties(wa_result, fake_item_factory, response_data):
    result = Result(wa_result, response_data, fake_item_factory)
    assert result.has_next is True
    assert result.has_previous is True
    assert result.before == "abc"
    assert result.after == "xyz"


def test_result_empty(wa_result, fake_item_factory, response_data):
    result = Result(wa_result, response_data, fake_item_factory)
    empty = result.empty
    assert isinstance(empty, Result)
    assert list(empty) == []
    assert empty.has_next is True  # next_url is preserved
    assert empty.has_previous is False
    assert empty.before == "abc"


def test_result_next(wa_result, fake_item_factory, response_data):
    wa_result.api._make_request.return_value = {
        "data": [{"name": "Charlie"}],
        "paging": {},
    }

    result = Result(wa_result, response_data, fake_item_factory)
    next_result = result.next()
    wa_result.api._make_request.assert_called_once_with(
        method="GET", endpoint="/next-url"
    )
    assert list(next_result) == [{"name": "Charlie"}]


def test_next_result_when_no_next(wa_result, fake_item_factory, response_data):
    response_data["paging"]["next"] = None
    result = Result(wa_result, response_data, fake_item_factory)
    next_result = result.next()
    assert not next_result.has_next
    assert next_result._data == []


def test_previous_result(wa_result, fake_item_factory, response_data):
    wa_result.api._make_request.return_value = {
        "data": [{"name": "Dave"}],
        "paging": {},
    }

    result = Result(wa_result, response_data, fake_item_factory)
    prev_result = result.previous()
    wa_result.api._make_request.assert_called_once_with(
        method="GET", endpoint="/prev-url"
    )
    assert list(prev_result) == [{"name": "Dave"}]


def test_previous_result_when_no_previous(wa_result, fake_item_factory, response_data):
    response_data["paging"]["previous"] = None
    result = Result(wa_result, response_data, fake_item_factory)
    prev_result = result.previous()
    assert not prev_result.has_previous
    assert prev_result._data == []


def test_all_on_first_page(wa_result, fake_item_factory, pages):
    first_page, second_page, third_page = pages
    wa_result.api._make_request.side_effect = [second_page, third_page]
    first_result = Result(wa_result, first_page, fake_item_factory)
    all_results = first_result.all()
    wa_result.api._make_request.assert_has_calls(
        [call(method="GET", endpoint="/page2"), call(method="GET", endpoint="/page3")],
        any_order=False,
    )
    assert all_results == [{"name": "Zed"}, {"name": "Yara"}, {"name": "Xander"}]


def test_all_on_middle_page(wa_result, fake_item_factory, pages):
    first_page, second_page, third_page = pages
    wa_result.api._make_request.side_effect = [first_page, third_page]
    second_result = Result(wa_result, second_page, fake_item_factory)
    all_results = second_result.all()
    wa_result.api._make_request.assert_has_calls(
        [
            call(method="GET", endpoint="/page1"),
            call(method="GET", endpoint="/page3"),
        ],
        any_order=False,
    )
    assert all_results == [{"name": "Zed"}, {"name": "Yara"}, {"name": "Xander"}]


def test_all_on_last_page(wa_result, fake_item_factory, pages):
    first_page, second_page, third_page = pages
    wa_result.api._make_request.side_effect = [second_page, first_page]
    third_result = Result(wa_result, third_page, fake_item_factory)
    all_results = third_result.all()
    wa_result.api._make_request.assert_has_calls(
        [
            call(method="GET", endpoint="/page2"),
            call(method="GET", endpoint="/page1"),
        ],
        any_order=False,
    )
    assert all_results == [{"name": "Zed"}, {"name": "Yara"}, {"name": "Xander"}]


def test_repr(wa_result, fake_item_factory, response_data):
    result = Result(wa_result, response_data, fake_item_factory)
    r = repr(result)
    assert "Result" in r
    assert "has_next=True" in r
    assert "has_previous=True" in r



import dataclasses

import pytest

from pywa.types import CallbackData


def test_callback_class_uniqueness():
    """Test the uniqueness of the callback data."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: str
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: str
        name: str

    assert User.__callback_id__ != Group.__callback_id__, (
        "The callback id must be unique for each child class."
    )


def test_callback_id_override():
    """Test the callback id override."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        __callback_id__ = "user"
        __callback_data_sep__ = "~"
        id: str
        name: str
        is_admin: bool

    assert User(id="1234", name="xxx", is_admin=True).to_str().startswith("user~")


def test_callback_class_not_empty():
    """Test if the callback data is empty."""
    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User1(CallbackData):
            pass

    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User2(CallbackData):
            id: str
            name: str
            is_admin: bool
            x: int

        @dataclasses.dataclass(slots=True, frozen=True)
        class UserChild(User2):
            pass


def test_callback_supported_types():
    """Test the supported types of the callback data."""
    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User(CallbackData):
            id: int
            name: str
            is_admin: bool
            x: dict
            y: list
            z: set


def test_data():
    """Test the callback data."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        __callback_id__ = "x"
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name="xxx", is_admin=True).to_str() == "x~1234~xxx~§"
    assert User(id=3456, name="yyy", is_admin=False).to_str() == "x~3456~yyy~"
    assert User.from_str("x~1234~xxx~§") == User(id=1234, name="xxx", is_admin=True)
    assert User.from_str("x~3456~yyy~") == User(id=3456, name="yyy", is_admin=False)

    with pytest.raises(ValueError):  # too few fields
        User.from_str("x~3456")

    with pytest.raises(ValueError):  # too many fields
        User.from_str("x~3456~yyy~§~")

    with pytest.raises(ValueError):  # invalid type
        User.from_str("x~x~yyy~§")


def test_data_sep():
    """Test the data separator override."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        __callback_id__ = "x"
        __callback_data_sep__ = "*"
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name="xxx", is_admin=True).to_str() == "x*1234*xxx*§"
    assert User(id=3456, name="yyy", is_admin=False).to_str() == "x*3456*yyy*"

    with pytest.raises(ValueError):
        User.from_str("x*3456*David*Lev*")

    try:
        User.from_str("x*3456*David~Lev*")
    except ValueError:
        pytest.fail("The data separator override does not work.")



from __future__ import annotations

"""This module contains the types related to system messages."""

__all__ = ["PhoneNumberChange", "IdentityChange", "SystemType", "Identity"]

import dataclasses

from pywa.types.system import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.system import (
    PhoneNumberChange as _UserChangedNumber,
    IdentityChange as _UserIdentityChanged,
)
from .base_update import BaseUserUpdateAsync  # noqa


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class PhoneNumberChange(BaseUserUpdateAsync, _UserChangedNumber):
    """
    A update received when a user changes their phone number on WhatsApp.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The type of the message (always ``MessageType.SYSTEM``).
        sys_type: The type of the system message (always ``SystemType.USER_CHANGED_NUMBER``).
        from_user: The user who changed their phone number. THIS IS THE OLD WA ID!
        timestamp: The timestamp when the message was arrived to WhatsApp servers (in UTC).
        old_wa_id: The old WhatsApp ID of the user.
        new_wa_id: The new WhatsApp ID of the user.
        body: The body of the system message (e.g., `John changed their phone number`).
    """


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class IdentityChange(BaseUserUpdateAsync, _UserIdentityChanged):
    """
    A message received when a user changes their profile information on WhatsApp.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The type of the message (always ``MessageType.SYSTEM``).
        from_user: The user who changed their profile information.
        timestamp: The timestamp when the message was arrived to WhatsApp servers (in UTC).
        body: The body of the system message (e.g., `John changed their profile information`).
        identity: The new identity of the user (see :class:`Identity`).
    """



"""This module contains updates related to user preferences in WhatsApp."""

from __future__ import annotations

__all__ = [
    "UserMarketingPreferences",
    "UserPreferenceCategory",
    "MarketingPreference",
]

import dataclasses
import datetime
import logging
from typing import TYPE_CHECKING

from .. import utils

from .base_update import BaseUserUpdate  # noqa
from .others import Metadata

if TYPE_CHECKING:
    from ..client import WhatsApp

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class UserMarketingPreferences(BaseUserUpdate):
    """
    Represents user marketing preferences updates from WhatsApp.

    - `'User preferences for marketing messages' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/send-message-templates#user-preferences-for-marketing-messages>`_.

    Attributes:
        id: The WhatsApp Business Account ID that the update was sent to.
        metadata: The metadata of the message (to which phone number it was sent).
        timestamp: The timestamp when the update was sent.
        from_user: The user who made the marketing preference change.
        value: The marketing preference chosen by the user, either ``stop`` or ``resume``.
        detail: A description of the marketing preference change (e.g. `User requested to stop marketing messages`).
        category: The category of the user preference, which is always ``marketing_messages``.
    """

    value: MarketingPreference
    detail: str
    category: UserPreferenceCategory

    _webhook_field = "user_preferences"

    @property
    def message_id_to_reply(self) -> str:
        """Raises an error because user preferences updates cannot be replied."""
        raise ValueError(
            "You cannot use `message_id_to_reply` to quote a user preference update."
        )

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> UserMarketingPreferences:
        prefs = (value := (entry := update["entry"][0])["changes"][0]["value"])[
            "user_preferences"
        ][0]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=entry["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            timestamp=datetime.datetime.fromtimestamp(
                int(prefs["timestamp"]),
                datetime.timezone.utc,
            ),
            from_user=client._usr_cls.from_dict(value["contacts"][0], client=client),
            value=MarketingPreference(prefs["value"]),
            detail=prefs["detail"],
            category=UserPreferenceCategory(prefs["category"]),
        )

    def __bool__(self):
        """Returns True if the user has not requested to stop receiving marketing messages."""
        return self.value != MarketingPreference.STOP


class MarketingPreference(utils.StrEnum):
    """
    The marketing preference chosen by the user.

    Attributes:
        STOP: The user has requested to stop receiving marketing messages.
        RESUME: The user has requested to resume receiving marketing messages.
    """

    _check_value = str.islower
    _modify_value = str.lower

    STOP = "stop"
    RESUME = "resume"

    UNKNOWN = "UNKNOWN"


class UserPreferenceCategory(utils.StrEnum):
    """
    The category of the user preference.

    Attributes:
        MARKETING_MESSAGES: User preferences for marketing messages.
    """

    _check_value = str.islower
    _modify_value = str.lower

    MARKETING_MESSAGES = "marketing_messages"

    UNKNOWN = "UNKNOWN"



"""The internal API for the WhatsApp client."""

import logging
from typing import Any, TYPE_CHECKING

import httpx

import pywa
from .errors import WhatsAppError

if TYPE_CHECKING:
    from .client import WhatsApp


_logger = logging.getLogger(__name__)


class GraphAPI:
    """Internal methods for the WhatsApp client. Do not use this class directly."""

    def __init__(
        self,
        token: str,
        session: httpx.Client,
        api_version: float,
    ):
        if session.headers.get("Authorization") is not None:
            raise ValueError(
                "You can't use the same httpx.Client for multiple WhatsApp instances!"
            )
        session.base_url = f"https://graph.facebook.com/v{api_version}"
        session.headers.update(
            {
                "Authorization": f"Bearer {token}",
                "User-Agent": f"PyWa/{pywa.__version__}",
            }
        )
        self._session = session

    def __str__(self) -> str:
        return f"GraphAPI(session={self._session})"

    def __repr__(self) -> str:
        return self.__str__()

    def _make_request(self, method: str, endpoint: str, **kwargs) -> dict | list:
        """
        Internal method to make a request to the WhatsApp Cloud API.

        Args:
            method: The HTTP method to use.
            endpoint: The endpoint to request.
            **kwargs: Additional arguments to pass to the request.

        Returns:
            The response JSON.

        Raises:
            WhatsAppError: If the request failed.
        """
        try:
            res = self._session.request(method=method, url=endpoint, **kwargs)
        except httpx.RequestError as e:
            e.add_note(
                "You may want to provide your own `httpx.Client` instance. e.g. `WhatsApp(session=httpx.Client(timeout=..., proxies=...))`. See https://www.python-httpx.org/api/#client for more information."
            )
            raise
        if res.status_code >= 400:
            raise WhatsAppError.from_dict(error=res.json()["error"], response=res)
        return res.json()

    def get_app_access_token(self, app_id: int, app_secret: str) -> dict[str, str]:
        """
        Get an access token for an app.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/facebook-login/guides/access-tokens/#apptokens>`_.

        Return example::

            {
                'access_token': 'xyzxyzxyz',
                'token_type': 'bearer'
            }


        Args:
            app_id: The ID of the app.
            app_secret: The secret of the app.

        Returns:
            The access token and its type.

        """
        return self._make_request(
            method="GET",
            endpoint="/oauth/access_token",
            params={
                "grant_type": "client_credentials",
                "client_id": app_id,
                "client_secret": app_secret,
            },
        )

    def set_app_callback_url(
        self,
        app_id: int,
        app_access_token: str,
        callback_url: str,
        verify_token: str,
        fields: tuple[str, ...],
    ) -> dict[str, bool]:
        """
        Set the callback URL for the webhook.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/reference/app/subscriptions>`_.

        Return example::

            {
                'success': True
            }

        Args:
            app_id: The ID of the app.
            app_access_token: The access token of the app (from ``get_app_access_token``).
            callback_url: The URL to set.
            verify_token: The verify token to challenge the webhook with.
            fields: The fields to subscribe to.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{app_id}/subscriptions",
            params={
                "object": "whatsapp_business_account",
                "callback_url": callback_url,
                "verify_token": verify_token,
                "fields": ",".join(fields),
                "access_token": app_access_token,
            },
        )

    def set_waba_alternate_callback_url(
        self,
        waba_id: str,
        callback_url: str,
        verify_token: str,
    ) -> dict[str, bool]:
        """
        Set an alternate callback URL on a WABA.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/embedded-signup/webhooks/override#set-waba-alternate-callback>`_.

        Return example::

            {
                'success': True
            }

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            callback_url: The URL to set.
            verify_token: The verify token to challenge the webhook with.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{waba_id}/subscribed_apps",
            json={
                "override_callback_uri": callback_url,
                "verify_token": verify_token,
            },
        )

    def get_waba_subscribed_apps(self, waba_id: str) -> dict[str, list[dict]]:
        """
        Get the subscribed apps of a WABA.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/embedded-signup/webhooks/override#get-waba-alternate-callback>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.

        Return example::

            {
                "data": [
                    {
                        "whatsapp_business_api_data": {
                            "link": "https://www.facebook.com/games/?app_id=12345",
                            "name": "My App",
                            "id": "12345"
                        },
                        "override_callback_uri": "https://example.com/wa_webhook"
                    },
                    {
                        "whatsapp_business_api_data": {
                            "link": "https://www.facebook.com/games/?app_id=67890",
                            "name": "My second app",
                            "id": "67890"
                        }
                    }
                ]
            }

        Returns:
            The subscribed apps of the WABA.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{waba_id}/subscribed_apps",
        )

    def delete_waba_alternate_callback_url(self, waba_id: str) -> dict[str, bool]:
        """
        Delete the callback URL of a WABA.

        Example:

            {
                'success': True
            }

        Args:
            waba_id: The ID of the WhatsApp Business Account.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{waba_id}/subscribed_apps",
        )

    def set_phone_alternate_callback_url(
        self,
        callback_url: str,
        verify_token: str,
        phone_id: str,
    ) -> dict[str, bool]:
        """
        Set an alternate callback URL on the business phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/embedded-signup/webhooks/override#set-phone-number-alternate-callback>`_.

        Args:
            phone_id: The ID of the phone number to set the callback URL on.
            callback_url: The URL to set.
            verify_token: The verify token to challenge the webhook with.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/",
            json={
                "webhook_configuration": {
                    "override_callback_uri": callback_url,
                    "verify_token": verify_token,
                }
            },
        )

    def delete_phone_alternate_callback_url(self, phone_id: str) -> dict[str, bool]:
        """
        Delete the alternate callback URL of a phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/embedded-signup/webhooks/override#delete-phone-number-alternate-callback>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to delete the callback URL from.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/",
            json={
                "webhook_configuration": {
                    "override_callback_uri": "",
                }
            },
        )

    def set_business_public_key(
        self,
        phone_id: str,
        public_key: str,
    ) -> dict[str, bool]:
        """
        Set the public key of the business.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/whatsapp-business-encryption/#set-business-public-key>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to set the public key on.
            public_key: The public key to set.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/whatsapp_business_encryption",
            data={"business_public_key": public_key},
        )

    def upload_media(
        self,
        phone_id: str,
        media: bytes,
        mime_type: str,
        filename: str,
    ) -> dict[str, str]:
        """
        Upload a media file to WhatsApp.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#upload-media>`_.

        Return example::

            {
              'id':'<MEDIA_ID>'
            }

        Args:
            phone_id: The ID of the phone number to upload the media to.
            media: media bytes or open(path, 'rb') object
            mime_type: The type of the media file
            filename: The name of the media file
        Returns:
            A dict with the ID of the uploaded media file.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/media",
            files={
                "file": (filename, media, mime_type),
                "messaging_product": (None, "whatsapp"),
                "type": (None, mime_type),
            },
        )

    def get_media_url(self, media_id: str) -> dict:
        """
        Get the URL of a media file.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#retrieve-media-url>`_.

        Return example::

            {
                'url': 'https://lookaside.fbsbx.com/whatsapp_business/attachments/?mid=645645&ext=54353&hash=xxx-xx',
                'mime_type': 'image/jpeg',
                'sha256': '73298ec14751fhfonf4wfxxxf52f4fb031db3892ff5',
                'file_size': 61901,
                'id': '137524587935346',
                'messaging_product': 'whatsapp'
            }

        Args:
            media_id: The ID of the media file.

        Returns:
            A dict with the URL and other info about the media file.
        """
        return self._make_request(method="GET", endpoint=f"/{media_id}")

    def get_media_bytes(
        self,
        media_url: str,
        **httpx_kwargs,
    ) -> tuple[bytes, str | None]:
        """
        Get the bytes of a media file from WhatsApp servers.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#download-media>`_.

        Args:
            media_url: The URL of the media file (from ``get_media_url``).
            **httpx_kwargs: Additional arguments to pass to the httpx get request.

        Returns:
            The media file bytes and the MIME type (if available).
        """
        headers = self._session.headers.copy()
        res = self._session.get(media_url, headers=headers, **httpx_kwargs)
        res.raise_for_status()
        return res.content, res.headers.get("Content-Type")

    def delete_media(
        self, media_id: str, phone_number_id: str | None = None
    ) -> dict[str, bool]:
        """
        Delete a media file from WhatsApp servers.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#delete-media>`_.

        Return example::

            {'success': True}

        Args:
            media_id: The ID of the media file.
            phone_number_id: Business phone number ID. If included, the operation will only be processed if the ID matches the ID of the business phone number that the media was uploaded on.

        Returns:
            True if the media file was deleted successfully, False otherwise.
        """
        params = {"phone_number_id": phone_number_id} if phone_number_id else None
        return self._make_request(
            method="DELETE", endpoint=f"/{media_id}", params=params
        )

    def send_raw_request(self, method: str, endpoint: str, **kwargs) -> Any:
        """
        Send a raw request to WhatsApp Cloud API.

        - Use this method if you want to send a request that is not yet supported by pywa.
        - Every request will automatically include the ``Authorization`` and ``Content-Type`` headers. you can override them by passing them in ``kwargs`` (headers={...}).

        Args:
            method: The HTTP method to use (e.g. ``POST``, ``GET``, etc.).
            endpoint: The endpoint to send the message to (e.g. ``/{phone_id}/messages/``).
            **kwargs: Additional arguments to send with the request (e.g. ``json={...}, headers={...}``).

        Example:

            >>> wa = WhatsApp(...)
            >>> wa.api.send_raw_request(
            ..   method="POST",
            ..   endpoint="/{phone_id}/messages",
            ..   json={"to": "1234567890", "type": "text", "text": {"body": "Hello, World!"}}
            .. )
            {
                'messaging_product': 'whatsapp',
                'contacts': [{'input': '1234567890', 'wa_id': '1234567890'}],
                'messages': [{'id': 'wamid.XXXXXXXXXXXXXXXX=='}]
            }

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method=method,
            endpoint=endpoint,
            **kwargs,
        )

    def send_message(
        self,
        sender: str,
        to: str,
        typ: str,
        msg: dict[str, str | list[str]] | tuple[dict],
        reply_to_message_id: str | None = None,
        biz_opaque_callback_data: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send a message to a WhatsApp user.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/messages>`_.

        Args:
            sender: The phone id to send the message from.
            to: The phone number to send the message to.
            typ: The type of the message (e.g. ``text``, ``image``, etc.).
            msg: The message object to send.
            reply_to_message_id: The ID of the message to reply to.
            biz_opaque_callback_data: The tracker to send with the message.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        data = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": to,
            "type": typ,
            typ: msg,
        }
        if reply_to_message_id:
            data["context"] = {"message_id": reply_to_message_id}
        if biz_opaque_callback_data:
            data["biz_opaque_callback_data"] = biz_opaque_callback_data
        return self._make_request(
            method="POST",
            endpoint=f"/{sender}/messages",
            json=data,
        )

    def send_marketing_message(
        self,
        sender: str,
        to: str,
        template: dict[str, str | list[str]],
        reply_to_message_id: str | None = None,
        message_activity_sharing: bool | None = None,
        biz_opaque_callback_data: str | None = None,
    ) -> dict[str, dict | list]:
        """
        Send marketing template messages via MM Lite API.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/marketing-messages-lite-api/sending-messages#sending-messages>`_.


        Args:
            sender: The phone id to send the message from.
            to: The phone number to send the message to.
            template: The template object to send.
            reply_to_message_id: The ID of the message to reply to.
            message_activity_sharing: Toggles on / off sharing message activities (e.g. message read) for that specific marketing message to Meta to help optimize marketing messages.
            biz_opaque_callback_data: The tracker to send with the message.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        body = {
            "messaging_product": "whatsapp",
            "recipient_type": "individual",
            "to": to,
            "type": "template",
            "template": template,
        }
        if reply_to_message_id:
            body["context"] = {"message_id": reply_to_message_id}
        if message_activity_sharing is not None:
            body["message_activity_sharing"] = message_activity_sharing
        if biz_opaque_callback_data:
            body["biz_opaque_callback_data"] = biz_opaque_callback_data
        return self._make_request(
            method="POST",
            endpoint=f"/{sender}/marketing_messages",
            json=body,
        )

    def register_phone_number(
        self,
        phone_id: str,
        pin: str,
        data_localization_region: str = None,
    ) -> dict[str, bool]:
        """
        Register a phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/registration#register>`_.

        Return example:
            {
                'success': True,
            }

        Args:
            phone_id: The ID of the phone number to register.
            pin: The pin to register the phone number with.
            data_localization_region: The region to localize the data (Value must be a 2-letter ISO 3166 country code (e.g. IN) indicating the country where you want data-at-rest to be stored).

        Returns:
            The success of the operation.
        """

        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/register",
            json={
                "messaging_product": "whatsapp",
                "pin": pin,
                **(
                    {"data_localization_region": data_localization_region}
                    if data_localization_region
                    else {}
                ),
            },
        )

    def deregister_phone_number(
        self,
        phone_id: str,
    ) -> dict[str, bool]:
        """
        Deregister a phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/registration/#deregister>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to deregister.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/deregister",
        )

    def mark_message_as_read(
        self,
        phone_id: str,
        message_id: str,
    ) -> dict[str, bool]:
        """
        Mark a message as read.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/mark-message-as-read>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number that message belongs to.
            message_id: The ID of the message to mark as read.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/messages",
            json={
                "messaging_product": "whatsapp",
                "status": "read",
                "message_id": message_id,
            },
        )

    def get_waba_info(
        self,
        waba_id: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict[str, Any]:
        """
        Get the WhatsApp Business Account information.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/reference/whats-app-business-account>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            fields: The fields to get. If None, all available fields will be returned.

        Returns:
            The WhatsApp Business Account information.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{waba_id}",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def get_business_phone_number(
        self,
        phone_id: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict[str, Any]:
        """
        Get the business phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/reference/whats-app-business-account-to-number-current-status/>`_.

        Return example::

            {
                'verified_name': 'Test Number',
                'code_verification_status': 'NOT_VERIFIED',
                'display_phone_number': '+1 555-096-7852',
                'quality_rating': 'GREEN',
                'platform_type': 'CLOUD_API',
                'throughput': {'level': 'STANDARD'},
                'id': '277321005464405'

            }
        Args:
            phone_id: The ID of the phone number to get.
            fields: The fields to get.

        Returns:
            The business phone number.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def get_business_phone_numbers(
        self,
        waba_id: str,
        fields: tuple[str, ...] | None = None,
        pagination: dict[str, str] | None = None,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Get business phone numbers.

        Return example::

            {
                'data': [
                    {
                        'verified_name': 'Test Number',
                        'code_verification_status': 'NOT_VERIFIED',
                        'display_phone_number': '+1 555-096-7852',
                        'quality_rating': 'GREEN',
                        'platform_type': 'CLOUD_API',
                        'throughput': {'level': 'STANDARD'},
                        'id': '277321005464405'
                    },
                    ...
                ]
            }

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            fields: The fields to get.
            pagination: The pagination of the API.

        Returns:
            The business phone numbers.
        """
        endpoint = f"/{waba_id}/phone_numbers"
        params = {
            k: v
            for k, v in {
                "fields": ",".join(fields) if fields else None,
            }.items()
            if v
        } | (pagination or {})
        return self._make_request(
            method="GET",
            endpoint=endpoint,
            params=params,
        )

    def get_business_phone_number_settings(
        self,
        phone_id: str,
        fields: tuple[str, ...] | None = None,
        include_sip_credentials: bool | None = None,
    ) -> dict[str, Any]:
        """
        Get the business phone number settings.

        Args:
            phone_id: The ID of the phone number to get.
            include_sip_credentials: Whether to include SIP credentials in the response.
            fields: The fields to get. If None, all available fields will be returned.

        Returns:
            The business phone number settings.
        """
        params = {
            "fields": ",".join(fields) if fields else None,
        }
        if include_sip_credentials is not None:
            params["include_sip_credentials"] = include_sip_credentials
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/settings",
            params=params,
        )

    def update_business_phone_number_settings(
        self,
        phone_id: str,
        settings: dict[str, Any],
    ) -> dict[str, bool]:
        """
        Update the business phone number settings.

        Args:
            phone_id: The ID of the phone number to update.
            settings: The settings to update.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/settings",
            json=settings,
        )

    def update_conversational_automation(
        self,
        phone_id: str,
        enable_welcome_message: bool | None = None,
        prompts: tuple[dict] | None = None,
        commands: str | None = None,
    ) -> dict[str, bool]:
        """
        Update the conversational automation settings.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers/conversational-components/#configuring-via-the-api>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to update.
            enable_welcome_message: Enable the welcome message.
            prompts: The prompts (ice breakers) to set.
            commands: The commands to set.

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/conversational_automation",
            json={
                k: v
                for k, v in {
                    "enable_welcome_message": enable_welcome_message,
                    "prompts": prompts,
                    "commands": commands,
                }.items()
                if v is not None
            },
        )

    def update_display_name(
        self,
        phone_id: str,
        new_display_name: str,
    ) -> dict[str, bool]:
        """
        Update the display name of the business phone number.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/phone-numbers#updating-display-name-via-api>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to update.
            new_display_name: The new display name to set.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}",
            json={
                "new_display_name": new_display_name,
                "messaging_product": "whatsapp",
            },
        )

    def get_business_profile(
        self,
        phone_id: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict[str, list[dict[str, str | list[str]]]]:
        """
        Get the business profile.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/business-profiles/#get-business-profile>`_.

        Return example::

            {
              "data": [{
                "about": "ABOUT",
                "address": "ADDRESS",
                "description": "DESCRIPTION",
                "email": "EMAIL",
                "messaging_product": "whatsapp",
                "profile_picture_url": "https://URL",
                "websites": [
                   "https://WEBSITE-1",
                   "https://WEBSITE-2"
                 ],
                "vertical": "INDUSTRY",
              }]
            }

        Args:
            phone_id: The ID of the phone number to get.
            fields: The fields to get.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/whatsapp_business_profile",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def update_business_profile(
        self, phone_id: str, data: dict[str, str | list[str]]
    ) -> dict[str, bool]:
        """
        Update the business profile.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/business-profiles/#update-business-profile>`_.

        Args:
            phone_id: The ID of the phone number to update.
            data: The data to update the business profile with.

        Return example::

            {
                "success": True
            }
        """
        data.update(messaging_product="whatsapp")
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/whatsapp_business_profile",
            json=data,
        )

    def get_commerce_settings(
        self, phone_id: str, fields: tuple[str, ...] | None = None
    ) -> dict:
        """
        Get the commerce settings of the business catalog.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/sell-products-and-services/set-commerce-settings/#get-commerce-settings>`_.

        Return example::

            {
              "data": [
                {
                  "is_cart_enabled": True,
                  "is_catalog_visible": True,
                  "id": "727705352028726"
                }
              ]
            }

        Args:
            phone_id: The ID of the phone number to get.
            fields: The fields to get. If None, all fields will be returned.

        Returns:
            The commerce settings of the business catalog.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/whatsapp_commerce_settings",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def update_commerce_settings(
        self,
        phone_id: str,
        data: dict,
    ) -> dict[str, bool]:
        """
        Change the commerce settings of the business catalog.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/sell-products-and-services/set-commerce-settings>`_.

        Args:
            phone_id: The ID of the phone number to update.
            data: The data to update the commerce settings with.

        Return example::

            {
              "success": True
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/whatsapp_commerce_settings",
            params=data,
        )

    def create_template(
        self,
        waba_id: str,
        template: dict,
    ) -> dict:
        """
        Create a message template.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#creating-templates>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            template: The template to create.

        Return example::

            {
                "id": "594425479261596",
                "status": "PENDING",
                "category": "MARKETING"
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{waba_id}/message_templates",
            json=template,
        )

    def get_template(
        self,
        template_id: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict[str, Any]:
        """
        Get a message template.

        Example::

            {
              "name": "seasonal_promotion_text_only",
              "status": "APPROVED",
              "id": "564750795574598"
            }

        Args:
            template_id: The ID of the template to get.
            fields: The fields to get. If None, all fields will be returned.

        Returns:
            The template data.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{template_id}",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def get_templates(
        self,
        waba_id: str,
        fields: tuple[str, ...] | None = None,
        filters: dict[str, Any] | None = None,
        summary_fields: tuple[str, ...] | None = None,
        pagination: dict[str, str] | None = None,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Get a list of message templates.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#retrieve-templates>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            fields: The fields to get. If None, all fields will be returned.
            filters: Filters to apply to the templates. e.g. {"status": "APPROVED"}.
            summary_fields: Aggregated information a such as counts.
            pagination: Pagination parameters to apply to the templates.

        Returns:
            A dict with the templates data.
        """
        params = {
            "fields": ",".join(fields) if fields else None,
            **(filters if filters else {}),
            **(pagination if pagination else {}),
            **({"summary": ",".join(summary_fields)} if summary_fields else {}),
        }
        return self._make_request(
            method="GET",
            endpoint=f"/{waba_id}/message_templates",
            params={k: v for k, v in params.items() if v is not None},
        )

    def update_template(
        self,
        template_id: str,
        template: dict,
    ) -> dict[str, bool]:
        """
        Update a message template.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#edit-a-message-template>`_.

        Args:
            template_id: The ID of the template to update.
            template: The template data to update.

        Returns:
            A dict with the success status of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{template_id}",
            json=template,
        )

    def delete_template(
        self,
        waba_id: str,
        template_name: str,
        template_id: str | None = None,
    ) -> dict[str, bool]:
        """
        Delete a message template.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#deleting-templates>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            template_name: The name of the template to delete.
            template_id: The ID of the template to delete. If not provided, the template will be deleted by name.

        Returns:
            A dict with the success status of the operation.
        """
        return self._make_request(
            method="DELETE",
            endpoint=f"/{waba_id}/message_templates",
            params={
                "name": template_name,
                **({"hsm_id": template_id} if template_id else {}),
            },
        )

    def compare_templates(
        self,
        template_id: str,
        template_ids: tuple[str, ...],
        start: str,
        end: str,
    ) -> dict:
        """
        Compare a template with other templates.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/template-comparison>`_.

        Args:
            template_id: The ID of the template to compare.
            template_ids: The IDs of the templates to compare with.
            start: TUNIX timestamp indicating start of timeframe.
            end: UNIX timestamp indicating end of timeframe.

        Returns:
            A dict with the comparison results.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{template_id}/compare",
            params={
                "template_ids": ",".join(template_ids),
                "start": start,
                "end": end,
            },
        )

    def migrate_templates(
        self,
        dest_waba_id: str,
        source_waba_id: str,
        page_number: int | None = None,
    ) -> dict:
        """
        Migrate templates from one WABA to another.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/template-migration>`_.

        Args:
            dest_waba_id: The ID of the destination WhatsApp Business Account.
            source_waba_id: The ID of the source WhatsApp Business Account.
            page_number: Indicates amount of templates to migrate as sets of 500. Zero-indexed. For example, to migrate 1000 templates, send one request with this value set to 0 and another request with this value set to 1, in parallel.

        Returns:
            A dict with the migration results.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{dest_waba_id}/migrate_message_templates",
            params={
                "source_waba_id": source_waba_id,
                **({"page_number": page_number} if page_number is not None else {}),
            },
        )

    def unpause_template(self, template_id: str) -> dict[str, bool | str]:
        """
        Unpause a message template.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/reference/whats-app-business-hsm/unpause/>`_.

        Args:
            template_id: The ID of the template to unpause.

        Returns:
            A dict with the success status of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{template_id}/unpause",
        )

    def upsert_message_templates(
        self,
        waba_id: str,
        template: dict,
    ) -> dict:
        """
        Bulk update or create authentication templates in multiple languages that include or exclude the optional security and expiration warnings.

        If a template already exists with a matching name and language, the template will be updated with the contents of the request, otherwise, a new template will be created.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/authentication-templates#bulk-management>`_.

        Return example::

            {
                "data": [
                    {
                        "id": "123456789012345",
                        "language": "en_US",
                        "status": "APPROVED"
                    },
                    ...
                ],

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            template: A dict containing the template to create or update to all languages listed in the template.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{waba_id}/upsert_message_templates",
            json=template,
        )

    def create_flow(
        self,
        waba_id: str,
        name: str,
        categories: tuple[str, ...],
        clone_flow_id: str | None = None,
        endpoint_uri: str | None = None,
        flow_json: str = None,
        publish: bool = None,
    ) -> dict[str, str]:
        """
        Create or clone a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#create>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            name: The name of the flow.
            categories: The categories of the flow.
            clone_flow_id: The ID of the flow to clone.
            endpoint_uri: The endpoint URI of the flow.
            flow_json: Flow's JSON encoded as string.
            publish: Whether to publish the flow. Only works if ``flow_json`` is also provided with valid Flow JSON.

        Return example::

            {
               "id": "<Flow-ID>"
               "success": True,
               "validation_errors": [
                {
                  "error": "INVALID_PROPERTY_VALUE" ,
                  "error_type": "FLOW_JSON_ERROR",
                  "message": "Invalid value found for property 'type'.",
                  "line_start": 10,
                  "line_end": 10,
                  "column_start": 21,
                  "column_end": 34,
                  "pointers": [
                   {
                     "line_start": 10,
                     "line_end": 10,
                     "column_start": 21,
                     "column_end": 34,
                     "path": "screens [0].layout.children[0].type"
                   }
                  ]
                }
              ]
            }
        """
        data = {
            "name": name,
            "categories": categories,
            **{
                k: v
                for k, v in {
                    "clone_flow_id": clone_flow_id,
                    "endpoint_uri": endpoint_uri,
                    "flow_json": flow_json,
                    "publish": publish,
                }.items()
                if v is not None
            },
        }
        return self._make_request(
            method="POST",
            endpoint=f"/{waba_id}/flows",
            json=data,
        )

    def update_flow_metadata(
        self,
        flow_id: str,
        name: str | None = None,
        categories: tuple[str, ...] | None = None,
        endpoint_uri: str | None = None,
        application_id: int | None = None,
    ) -> dict[str, bool]:
        """
        Update the metadata of a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#update>`_.

        Args:
            flow_id: The ID of the flow.
            name: The name of the flow.
            categories: The categories of the flow.
            endpoint_uri: The endpoint URI of the flow.
            application_id: The ID of the application.

        Return example::

            {
              "success": True
            }
        """
        data = {
            **({"name": name} if name else {}),
            **({"endpoint_uri": endpoint_uri} if endpoint_uri else {}),
            **({"categories": categories} if categories else {}),
            **({"application_id": application_id} if application_id else {}),
        }
        return self._make_request(
            method="POST",
            endpoint=f"/{flow_id}",
            json=data,
        )

    def update_flow_json(self, flow_id: str, flow_json: str) -> dict:
        """
        Update the JSON of a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#update-json>`_.

        Args:
            flow_id: The ID of the flow.
            flow_json: The JSON of the flow.

        Return example::

            {
              "success": true,
              "validation_errors": [
                {
                  "error": "INVALID_PROPERTY",
                  "error_type": "JSON_SCHEMA_ERROR",
                  "message": "The property \"initial-text\" cannot be specified at \"$root/screens/0/layout/children/2/children/0\".",
                  "line_start": 46,
                  "line_end": 46,
                  "column_start": 17,
                  "column_end": 30
                }
              ]
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{flow_id}/assets",
            files={
                "file": ("flow.json", flow_json, "application/json"),
                "name": (None, "flow.json"),
                "asset_type": (None, "FLOW_JSON"),
                "messaging_product": (None, "whatsapp"),
            },
        )

    def publish_flow(
        self,
        flow_id: str,
    ) -> dict[str, bool]:
        """
        Publish a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#publish>`_.

        Args:
            flow_id: The ID of the flow.

        Return example::

            {
              "success": true
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{flow_id}/publish",
        )

    def delete_flow(
        self,
        flow_id: str,
    ) -> dict[str, bool]:
        """
        Delete a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#delete>`_.

        Args:
            flow_id: The ID of the flow.

        Return example::

            {
              "success": true
            }
        """

        return self._make_request(
            method="DELETE",
            endpoint=f"/{flow_id}",
        )

    def deprecate_flow(
        self,
        flow_id: str,
    ) -> dict[str, bool]:
        """
        Deprecate a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#deprecate>`_.

        Args:
            flow_id: The ID of the flow.

        Return example::

            {
              "success": true
            }
        """

        return self._make_request(
            method="POST",
            endpoint=f"/{flow_id}/deprecate",
        )

    def get_flow(
        self,
        flow_id: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict[str, Any]:
        """
        Get a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#details>`_.

        Args:
            flow_id: The ID of the flow.
            fields: The fields to get.

        Return example::

            {
              "id": "<Flow-ID>",
              "name": "<Flow-Name>",
              "status": "DRAFT",
              "categories": [ "LEAD_GENERATION" ],
              "validation_errors": [],
              "json_version": "3.0",
              "data_api_version": "3.0",
              "endpoint_uri": "https://example.com",
              "preview": {
                "preview_url": "https://business.facebook.com/wa/manage/flows/55000..../preview/?token=b9d6.....",
                "expires_at": "2023-05-21T11:18:09+0000"
              },
              "whatsapp_business_account": {
                ...
              },
              "application": {
                ...
              }
            }
        """
        endpoint = f"/{flow_id}"
        if fields:
            endpoint += f"?fields={','.join(fields)}"
        return self._make_request(
            method="GET",
            endpoint=endpoint,
        )

    def get_flows(
        self,
        waba_id: str,
        fields: tuple[str, ...] | None = None,
        pagination: dict[str, str] | None = None,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Get all flows.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#list>`_.

        Args:
            waba_id: The ID of the WhatsApp Business Account.
            fields: The fields to get.
            pagination: The pagination of the API.

        Return example::

            {
              "data": [
                {
                  "id": "<Flow-ID>",
                  "name": "<Flow-Name>",
                  "status": "DRAFT",
                  "categories": [ "LEAD_GENERATION" ]
                },
                {
                    "id": "<Flow-ID>",
                    "name": "<Flow-Name>",
                    "status": "DRAFT",
                    "categories": [ "LEAD_GENERATION" ]
                }
              ]
            }
        """
        endpoint = f"/{waba_id}/flows"
        params = {
            k: v
            for k, v in {
                "fields": ",".join(fields) if fields else None,
            }.items()
            if v
        } | (pagination or {})
        return self._make_request(
            method="GET",
            endpoint=endpoint,
            params=params,
        )

    def get_flow_assets(
        self,
        flow_id: str,
        pagination: dict[str, str] | None = None,
    ) -> dict[str, list | dict]:
        """
        Get all assets of a flow.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi#asset-list>`_.

        Args:
            flow_id: The ID of the flow.
            pagination: The pagination of the API.

        Return example::

            {
              "data": [
                {
                  "name": "flow.json",
                  "asset_type": "FLOW_JSON",
                  "download_url": "https://scontent.xx.fbcdn.net/m1/v/t0.57323-24/An_Hq0jnfJ..."
                }
              ],
              "paging": {
                "cursors": {
                  "before": "QVFIU...",
                  "after": "QVFIU..."
                }
              }
            }
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{flow_id}/assets?fields=name,asset_type,download_url",
            params=pagination,
        )

    def migrate_flows(
        self,
        dest_waba_id: str,
        source_waba_id: str,
        source_flow_names: tuple[str, ...],
    ) -> dict[str, list[dict[str, str]]]:
        """
        Migrate Flows from one WhatsApp Business Account (WABA) to another.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowsapi/#migrate>`_.

        Migration doesn't move the source Flows, it creates copies of them with the same names in the destination WABA.

        - You can specify specific Flow names to migrate, or choose to migrate all Flows in source WABA.
        - Flows can only be migrated between WABAs owned by the same Meta business.
        - If a Flow exists with the same name in the destination WABA, it will be skipped and the API will return an error message for that Flow. Other Flows in the same request will be copied.
        - The migrated Flow will be published if the original Flow is published, otherwise it will be in draft state.
        - New Flows under destination WABA will have new Flow IDs.


        Return example::

            {
              "migrated_flows": [
                {
                  "source_name": "appointment-booking",
                  "source_id": "1234",
                  "migrated_id": "5678"
                }
              ],
              "failed_flows": [
                {
                  "source_name": "lead-gen",
                  "error_code": "4233041",
                  "error_message": "Flows Migration Error: Flow with the same name exists in destination WABA."
                }
              ]
            }

        Args:
            dest_waba_id: Destination WhatsApp Business Account ID.
            source_waba_id: Source WhatsApp Business Account ID.
            source_flow_names: Tuple of specific Flow names to migrate. If not specified, it will migrate all flows in source WABA. Only 100 Flows can be migrated in a request.

        Returns:
            Result of the migration request.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{dest_waba_id}/migrate_flows?source_waba_id={source_waba_id}&source_flow_names={','.join(source_flow_names)}",
        )

    def create_qr_code(
        self, phone_id: str, prefilled_message: str, generate_qr_image: str
    ) -> dict:
        """
        Create a QR code.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/qr-codes/#create-qr-code>`_.

        Args:
            phone_id: The ID of the phone number to create the QR code on.
            prefilled_message: The prefilled message to set.
            generate_qr_image: PNG or SVG

        Return example::

            {
              "code": "4O4YGZEG3RIVE1",
              "prefilled_message": "Cyber Monday 1",
              "deep_link_url": "https://wa.me/message/4O4YGZEG3RIVE1",
              "qr_image_url": "https://scontent-iad3-2.xx.fbcdn.net/..."
            }
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/message_qrdls",
            json={
                "prefilled_message": prefilled_message,
                "generate_qr_image": generate_qr_image,
            },
        )

    def get_qr_code(
        self,
        phone_id: str,
        code: str,
        fields: tuple[str, ...] | None = None,
    ) -> dict:
        """
        Get a QR code.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/qr-codes/#get-qr-code>`_.

        Args:
            phone_id: The ID of the phone number to get the QR code from.
            code: The code of the QR code.
            fields: The fields to get. If None, default fields will be returned.

        Return example::

            {
              "data": [
                {
                  "code": "4O4YGZEG3RIVE1",
                  "prefilled_message": "Cyber Monday",
                  "deep_link_url": "https://wa.me/message/4O4YGZEG3RIVE1"
                }
              ]
            }
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/message_qrdls/{code}",
            params={"fields": ",".join(fields)} if fields else None,
        )

    def get_qr_codes(
        self,
        phone_id: str,
        fields: tuple[str, ...] | None = None,
        pagination: dict[str, str] | None = None,
    ) -> dict:
        """
        Get all QR codes.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/qr-codes/#get-qr-codes>`_.

        Return example::

            {
              "data": [
                {
                  "code": "4O4YGZEG3RIVE1",
                  "prefilled_message": "Cyber Monday 1",
                  "deep_link_url": "https://wa.me/message/4O4YGZEG3RIVE1",
                  "qr_image_url": "https://scontent-iad3-2.xx.fbcdn.net/..."
                }
              ],
                "paging": {
                    "cursors": {
                    "before": "QVFIU...",
                    "after": "QVFIU..."
                    }
                }
            }

        Args:
            phone_id: The ID of the phone number to get the QR codes from.
            fields: The fields to get. If None, default fields will be returned.
            pagination: The pagination parameters.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/message_qrdls",
            params={
                k: v
                for k, v in {
                    "fields": ",".join(fields) if fields else None,
                    **(pagination or {}),
                }.items()
                if v is not None
            },
        )

    def update_qr_code(
        self,
        phone_id: str,
        code: str,
        prefilled_message: str,
    ) -> dict:
        """
        Update a QR code.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/qr-codes/#update-qr-code>`_.

        Return example::

            {
              "code": "4O4YGZEG3RIVE1",
              "prefilled_message": "Cyber Tuesday",
              "deep_link_url": "https://wa.me/message/4O4YGZEG3RIVE1"
            }

        Args:
            phone_id: The ID of the phone number to update the QR code on.
            code: The code of the QR code.
            prefilled_message: The prefilled message to set.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/message_qrdls",
            json={"code": code, "prefilled_message": prefilled_message},
        )

    def delete_qr_code(self, phone_id: str, code: str):
        """
        Delete a QR code.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/business-management-api/qr-codes/#delete-qr-code>`_.

        Return example::

            {
              "success": True
            }

        Args:
            phone_id: The ID of the phone number to delete the QR code from.
            code: The code of the QR code.
        """
        return self._make_request(
            method="DELETE",
            endpoint=f"/{phone_id}/message_qrdls/{code}",
        )

    def set_indicator(
        self, phone_id: str, message_id: str, typ: str
    ) -> dict[str, bool]:
        """
        Set a typing indicator.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/typing-indicators>`_.

        Return example::

            {
                'success': True
            }

        Args:
            phone_id: The ID of the phone number to set the typing indicator on.
            message_id: The ID of the message in the conversation to set the typing indicator on.
            typ: The type of indicator to set (currently only "text" is supported).

        Returns:
            The success of the operation.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/messages",
            json={
                "messaging_product": "whatsapp",
                "status": "read",
                "message_id": message_id,
                "typing_indicator": {
                    "type": typ,
                },
            },
        )

    def block_users(
        self,
        phone_id: str,
        users: tuple[str, ...],
    ) -> dict:
        """
        Block users from sending messages to the business.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/block-users#block-users-2>`_.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'block_users': {
                    'removed_users': [
                        {'input': '123456789', 'wa_id': '123456789'}
                    ]
                }
            }

        Args:
            phone_id: The ID of the phone number to block users on.
            users: The list of phone numbers/wa_ids to block.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{phone_id}/block_users",
            json={
                "messaging_product": "whatsapp",
                "block_users": [{"user": user} for user in users],
            },
        )

    def unblock_users(
        self,
        phone_id: str,
        users: tuple[str, ...],
    ) -> dict:
        """
        Unblock users from sending messages to the business.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/block-users#unblock-users>`_.

        Return example::

            {
                'messaging_product': 'whatsapp',
                'block_users': {
                    'removed_users': [
                        {'input': '123456789', 'wa_id': '123456789'}
                    ]
                }

        Args:
            phone_id: The ID of the phone number to unblock users on.
            users: The list of phone numbers/wa_ids to unblock.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="DELETE",
            endpoint=f"/{phone_id}/block_users",
            json={
                "messaging_product": "whatsapp",
                "block_users": [{"user": user} for user in users],
            },
        )

    def get_blocked_users(self, phone_id: str, pagination: dict | None = None) -> dict:
        """
        Get the list of blocked users.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/block-users#get-list-of-blocked-numbers>`_.

        Args:
            phone_id: The ID of the phone number to get the blocked users from.
            pagination: The pagination parameters.

        Return example::

            {
                'data': [
                    {'messaging_product': 'whatsapp', 'wa_id': '1234567890'}
                ],
                'paging': {
                    'cursors': {
                        'before': 'eyJvZAmZAzZAXQiOjAsIn',
                        'after': 'I6IjE3Mzc1ODA5MjUwMTk0NTIifQZDZD',
                    }
                }
            }

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/block_users",
            params=pagination,
        )

    def create_upload_session(
        self,
        app_id: str,
        file_name: str,
        file_length: int,
        file_type: str,
    ) -> dict[str, str]:
        """
        Create an upload session for a file.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/guides/upload#step-1>`_.

        Args:
            app_id: The ID of the app to create the upload session for.
            file_name: The name of the file to upload.
            file_length: File size in bytes
            file_type: The file's MIME type. Valid values are: ``application/pdf``, ``image/jpeg``, ``image/jpg``, ``image/png``, and ``video/mp4``.

        Returns:
            The ID of the upload session
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{app_id}/uploads?file_name={file_name}&file_length={file_length}&file_type={file_type}",
        )

    def upload_file(
        self,
        upload_session_id: str,
        file: bytes,
        file_offset: int = 0,
    ) -> dict[str, str]:
        """
        Upload a file to an upload session.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/guides/upload#step-2>`_.

        Args:
            upload_session_id: The ID of the upload session to upload the file to (This is the ID returned by the ``create_upload_session`` method).
            file: The file to upload (as bytes).
            file_offset: The offset in the file to start uploading from. you can use this to resume an interrupted upload (use `get_upload_session` to get the current offset).

        Returns:
            The file handle of the uploaded file.
        """
        return self._make_request(
            method="POST",
            endpoint=f"/{upload_session_id}",
            headers={
                "file_offset": str(file_offset),
            },
            data=file,
        )

    def get_upload_session(
        self,
        upload_session_id: str,
    ) -> dict[str, str | int]:
        """
        Get the status of an upload session (resuming an interrupted upload).

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/graph-api/guides/upload#resume-an-interrupted-upload>`_.

        Args:
            upload_session_id: The ID of the upload session.

        Returns:
            The status of the upload session (ID and current offset)
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{upload_session_id}",
        )

    def get_call_permissions(self, phone_id: str, user_wa_id: str) -> dict:
        """
        Get call permissions for a user.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-call-permissions>`_.

        Return example::

            {
              "messaging_product": "whatsapp",
              "permission": {
                "status": "temporary",
                "expiration_time": 1745343479
              },
              "actions": [
                {
                  "action_name": "send_call_permission_request",
                  "can_perform_action": True,
                  "limits": [
                    {
                      "time_period": "PT24H",
                      "max_allowed": 1,
                      "current_usage": 0,
                    },
                    {
                      "time_period": "P7D",
                      "max_allowed": 2,
                      "current_usage": 1,
                    }
                  ]
                },
                {
                  "action_name": "start_call",
                  "can_perform_action": False,
                  "limits": [
                    {
                      "time_period": "PT24H",
                      "max_allowed": 5,
                      "current_usage": 5,
                      "limit_expiration_time": 1745622600,
                    }
                  ]
                }
              }
            }

        Args:
            phone_id: The ID of the phone number to get call permissions for.
            user_wa_id: The WhatsApp ID of the user to check permissions for.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="GET",
            endpoint=f"/{phone_id}/call_permissions",
            params={"user_wa_id": user_wa_id},
        )

    def initiate_call(
        self,
        phone_id: str,
        to: str,
        sdp: dict[str, str],
        biz_opaque_callback_data: str | None = None,
    ) -> dict[str, str | bool]:
        """
        Initiate a call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#initiate-call>`_.

        Return example::

            {
              "messaging_product": "whatsapp",
              "calls" : [{
                 "id" : "wacid.ABGGFjFVU2AfAgo6V",
               }]
            }

        Args:
            phone_id: The ID of the phone number to initiate the call on.
            to: The number being called (callee).
            sdp: The SDP info of the device on the other end of the call. The SDP must be compliant with RFC 8866.
            biz_opaque_callback_data: An arbitrary string you can pass in that is useful for tracking and logging purposes.

        Returns:
            The response from the WhatsApp Cloud API containing the call ID.
        """
        return self._make_request(
            method="POST",
            endpoint=f"{phone_id}/calls",
            json={
                "messaging_product": "whatsapp",
                "to": to,
                "action": "connect",
                "session": sdp,
                **(
                    {"biz_opaque_callback_data": biz_opaque_callback_data}
                    if biz_opaque_callback_data
                    else {}
                ),
            },
        )

    def pre_accept_call(
        self, phone_id: str, call_id: str, sdp: dict[str, str] | None = None
    ) -> dict[str, str | bool]:
        """
        Pre-accept a call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#pre-accept-call>`_.

        Return example::

            {
                "messaging_product": "whatsapp",
                "success": True
            }

        Args:
            phone_id: The ID of the phone number to pre-accept the call on.
            call_id: The ID of the call to pre-accept.
            sdp: The SDP info of the device on the other end of the call. The SDP must be compliant with RFC 8866.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="POST",
            endpoint=f"{phone_id}/calls",
            json={
                "messaging_product": "whatsapp",
                "call_id": call_id,
                "action": "pre_accept",
                **({"session": sdp} if sdp else {}),
            },
        )

    def accept_call(
        self,
        phone_id: str,
        call_id: str,
        sdp: dict[str, str] | None = None,
        biz_opaque_callback_data: str | None = None,
    ) -> dict[str, str | bool]:
        """
        Accept a call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#accept-call>`_.

        Return example::

            {
                "messaging_product": "whatsapp",
                "success": True
            }

        Args:
            phone_id: The ID of the phone number to accept the call on.
            call_id: The ID of the call to accept.
            sdp: The SDP info of the device on the other end of the call. The SDP must be compliant with RFC 8866.
            biz_opaque_callback_data: An arbitrary string you can pass in that is useful for tracking and logging purposes.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="POST",
            endpoint=f"{phone_id}/calls",
            json={
                "messaging_product": "whatsapp",
                "call_id": call_id,
                "action": "accept",
                **({"session": sdp} if sdp else {}),
                **(
                    {"biz_opaque_callback_data": biz_opaque_callback_data}
                    if biz_opaque_callback_data
                    else {}
                ),
            },
        )

    def reject_call(self, phone_id: str, call_id: str) -> dict[str, bool]:
        """
        Reject a call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#reject-call>`_.

        Return example::

            {
                "messaging_product": "whatsapp",
                "success": True
            }

        Args:
            phone_id: The ID of the phone number to reject the call on.
            call_id: The ID of the call to reject.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="POST",
            endpoint=f"{phone_id}/calls",
            json={
                "messaging_product": "whatsapp",
                "call_id": call_id,
                "action": "reject",
            },
        )

    def terminate_call(self, phone_id: str, call_id: str) -> dict[str, bool]:
        """
        Terminate a call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#terminate-call>`_.

        Return example::

            {
                "messaging_product": "whatsapp",
                "success": True
            }

        Args:
            phone_id: The ID of the phone number to terminate the call on.
            call_id: The ID of the call to terminate.

        Returns:
            The response from the WhatsApp Cloud API.
        """
        return self._make_request(
            method="POST",
            endpoint=f"{phone_id}/calls",
            json={
                "messaging_product": "whatsapp",
                "call_id": call_id,
                "action": "terminate",
            },
        )



from pywa.utils import *  # noqa MUST BE IMPORTED FIRST
import httpx
from pywa.utils import _flow_request_media_decryptor


async def flow_request_media_decryptor(
    encrypted_media: dict[str, str | dict[str, str]],
    dl_session: httpx.AsyncClient | None = None,
) -> FlowRequestDecryptedMedia:
    """
    Decrypt the encrypted media file from the flow request.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/flowjson/components/media_upload#endpoint>`_.
    - Use the .decrypt_media() shorthand method of the :class:`FlowRequest` class instead.
    - This implementation requires ``cryptography`` to be installed. To install it, run ``pip3 install 'pywa[cryptography]'`` or ``pip3 install cryptography``.

    Example:

        >>> from pywa_async import WhatsApp, types
        >>> wa = WhatsApp(...)
        >>> @wa.on_flow_request("/media-upload")
        ... async def on_media_upload_request(_: WhatsApp, req: types.FlowRequest) -> types.FlowResponse | None:
        ...     dec = await req.decrypt_media(key="driver_license", index=0)
        ...     with open(dec.filename, "wb") as file:
        ...         file.write(dec.data)
        ...     return req.respond(...)

    Args:
        encrypted_media (dict): encrypted media data from the flow request (see example above).
        dl_session (httpx.AsyncClient): download session. Optional.

    Returns:
        tuple[str, str, bytes]
        - media_id (str): media ID
        - filename (str): media filename
        - decrypted_data (bytes): decrypted media file

    Raises:
        HTTPStatusError: If the request to the CDN URL fails.
        ValueError: If any of the hash verifications fail.
    """
    res = await (dl_session or httpx.AsyncClient()).get(encrypted_media["cdn_url"])
    res.raise_for_status()
    return FlowRequestDecryptedMedia(
        media_id=encrypted_media["media_id"],
        filename=encrypted_media["file_name"],
        data=_flow_request_media_decryptor(
            res.content, encrypted_media["encryption_metadata"]
        ),
    )



"""This module contains updates related to user preferences in WhatsApp."""

from __future__ import annotations

__all__ = [
    "UserMarketingPreferences",
    "UserPreferenceCategory",
    "MarketingPreference",
]

from pywa.types.user_preferences import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.user_preferences import (
    UserMarketingPreferences as _UserMarketingPreferences,
)  # noqa MUST BE IMPORTED FIRST

import dataclasses

from .base_update import BaseUserUpdateAsync  # noqa


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class UserMarketingPreferences(BaseUserUpdateAsync, _UserMarketingPreferences):
    """
    Represents user marketing preferences updates from WhatsApp.

    - `'User preferences for marketing messages' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/send-message-templates#user-preferences-for-marketing-messages>`_.

    Attributes:
        id: The WhatsApp Business Account ID that the update was sent to.
        metadata: The metadata of the message (to which phone number it was sent).
        timestamp: The timestamp when the update was sent.
        from_user: The user who made the marketing preference change.
        value: The marketing preference chosen by the user, either ``stop`` or ``resume``.
        detail: A description of the marketing preference change (e.g. `User requested to stop marketing messages`).
        category: The category of the user preference, which is always ``marketing_messages``.
    """
